# Tools for defining and running experiments
import asyncio
import collections
import datetime
import importlib
import logging
import numbers
import os
import pathlib
import sys
import threading
import time
import tkinter as tk
import warnings
from tkinter.filedialog import askopenfilename
from typing import Union

import numpy as np
import pandas as pd
import pykwalify.errors
from pandas.core.computation import expressions
from pykwalify.core import Core as YamlValidator
from ruamel.yaml import YAML

from empyric import variables as _variables
from empyric import adapters as _adapters
from empyric import graphics as _graphics
from empyric import instruments as _instruments
from empyric import routines as _routines
from empyric.adapters import AdapterError

from empyric.tools import convert_time, Clock, logger
from empyric.types import recast, Boolean, Toggle, Integer, Float, ON


class Experiment:
    """
    An iterable class which represents an experiment; iterates through any
    assigned routines, and retrieves and stores the values of all experiment
    variables. Each variable and routine is updated once per iteration, and each
    iteration blocks until all variables and routines are updated.

    The constructor take a `variables` argument in the form of a dictionary with the
    format {..., name: variable, ...}, which contains all of the variables controlled
    and monitored by the experiment. The optional `routines` argument is a dictionary
    of the form {..., name: routine, ...} containing any routines to run within
    the loop of the experiment. The optional `end` argument indicates when the
    experiment should end (i.e. raise `StopIteration` on subsequent call to `__next__`)
    , and can either be a number, a string of the form "[number] [time unit, e.g.
    seconds, minutes or hours]" or "with routines" to end the experiment after the
    last routine has ended.
    """

    # Possible statuses of an experiment
    READY = "Ready"  # Experiment is waiting to start
    RUNNING = "Running"  # Experiment is running
    HOLDING = "Holding"  # Routines are stopped, but measurements are ongoing
    STOPPED = "Stopped"  # Both routines and measurements are stopped
    TERMINATED = "Terminated"

    # Experiment has either finished or has been terminated by the user

    @property
    def status(self):
        return self._status

    @status.setter
    def status(self, status):
        prior_base_status = self._status.split(":")[0]
        new_base_status = status.split(":")[0]

        # Only allow change if the status is unlocked,
        # or if the base status is the same
        if not self.status_locked or new_base_status == prior_base_status:
            self._status = status

    @property
    def ready(self):
        return "Ready" in self.status

    @property
    def running(self):
        return "Running" in self.status

    @property
    def holding(self):
        return "Holding" in self.status

    @property
    def stopped(self):
        return "Stopped" in self.status

    @property
    def terminated(self):
        return "Terminated" in self.status

    def __init__(
            self,
            variables: dict,
            routines: dict = None,
            end: Union[numbers.Number, str, None] = None,
    ):
        self.variables = variables
        # dict of the form {..., name: variable, ...}

        # Used to block evaluation of expressions
        # until their dependee variables are evaluated
        for variable in self.variables.values():
            variable._eval_event = threading.Event()

        if routines:
            self.routines = routines
            # dictionary of the form {..., name: (variable_name, routine), ...}
        else:
            self.routines = {}

        if end:
            if type(end) is str:
                if end.lower() == "with routines":
                    self.end = max([routine.end for routine in routines.values()])
                else:
                    self.end = convert_time(end)
            elif isinstance(end, numbers.Number):
                self.end = end
            else:
                raise ValueError("invalid value for end keyword argument")
        else:
            self.end = np.inf

        self.clock = Clock()
        self.clock.start()

        self.timestamp = datetime.datetime.now().strftime("%Y%m%d-%H%M%S")

        self.state = pd.Series(
            name=None,
            data={**{"Time": None}, **{name: None for name in self.variables}},
            dtype=object,
        )
        self.data = pd.DataFrame(
            columns=["Time"] + list(variables.keys()), dtype=object
        )

        self._status = Experiment.READY
        self.status_locked = True
        # can only be unlocked by the start, hold, stop and terminate methods

        self.saved = []  # list of saved data entries

    def __next__(self):

        # Start the clock on first call
        if self.state.name is None:  # first step of the experiment
            self.start()
            self.status = Experiment.RUNNING

        # Update time
        self.state["Time"] = self.clock.time
        self.state.name = datetime.datetime.now()

        logger.info(f'Iterating experiment (t = {self.state["Time"]} s)')

        if self.stopped:
            return self.state

        # If the experiment is running, apply new settings to knobs
        # according to the routines (if there are any)
        if self.running:
            # Update each routine in its own thread
            threads = {}
            for name, routine in self.routines.items():
                threads[name] = threading.Thread(
                    target=self._update_routine, args=(name,)
                )
                threads[name].start()

            # Wait for all routine threads to finish
            for name, thread in threads.items():
                self.status = Experiment.RUNNING + f": executing {name}"
                thread.join()

            self.status = Experiment.RUNNING

        # Get all variable values if experiment is running or holding
        if self.running or self.holding:
            for variable in self.variables.values():
                variable._eval_event.clear()

            # Run each measure / get operation in its own thread
            threads = {}
            for name in self.variables:
                threads[name] = threading.Thread(
                    target=self._update_variable, args=(name,)
                )
                threads[name].start()

            base_status = self.status

            # Wait for all threads to finish
            for name, thread in threads.items():
                self.status = base_status + f": retrieving {name}"
                thread.join()

            self.status = base_status

            # Append new state to experiment data set
            with warnings.catch_warnings():
                warnings.simplefilter(action="ignore", category=FutureWarning)
                self.data.loc[self.state.name] = self.state

        # End the experiment, if the duration of the experiment has passed
        if self.clock.time > self.end:
            self.terminate()

        if self.terminated:
            raise StopIteration

        return self.state

    def __iter__(self):
        return self

    def _update_variable(self, name):
        """Retrieve and store a variable value"""

        try:
            if isinstance(self.variables[name], _variables.Expression):
                for symbol, dependee in self.variables[name].definitions.items():
                    if hasattr(dependee, "_eval_event"):
                        logger.debug(
                            f"Expression {name} is waiting for {symbol} to be evaluated"
                        )

                        dependee._eval_event.wait()

            try:
                value = self.variables[name].value
                logger.info(f"{name} evaluated to {value}")
            except AdapterError as adapter_error:
                value = None
                logger.warning(f"Unable to evaluate {name}: {adapter_error}")
            except ValueError as value_error:
                value = None
                logger.warning(f"Unable to evaluate {name}: {value_error}")

            self.variables[name]._eval_event.set()
            # unblock threads evaluating dependents

            if np.size(value) > 1:  # store array data as CSV files
                if np.any(value):
                    # only save non-empty arrays
                    dataframe = pd.DataFrame(
                        {name: value}, index=[self.state.name] * len(value)
                    )
                    path = name.replace(" ", "_") + "_"
                    path += self.state.name.strftime("%Y%m%d-%H%M%S") + ".csv"
                    dataframe.to_csv(path)

                    self.state[name] = os.path.abspath(path)
                else:
                    self.state[name] = None
            else:
                self.state[name] = value
        except Exception as err:
            self.terminate()
            raise err

    def _update_routine(self, name):
        """Update a routine according to the current state"""

        try:
            try:
                self.routines[name].update(self.state)
            except AdapterError as adapter_error:
                logger.warning(f"Unable to update routine {name}: {adapter_error}")
        except Exception as err:
            self.terminate()
            raise err

    def save(self, directory=None):
        """
        Save the experiment dataframe to a CSV file

        :param directory: (path) (optional) directory to save data to,
                          if different from working directory
        :return: None
        """

        base_status = self.status
        self.status = base_status + ": saving data"

        path = f"data_{self.timestamp}.csv"

        if directory:
            path = os.path.join(directory, path)

        logger.info(f"Saving experiment data to {path}")

        if not os.path.exists(path):
            # if this is a new file, write column headers
            with open(path, "a") as data_file:
                data_file.write("," + ",".join(self.data.columns) + "\n")

        unsaved = np.setdiff1d(self.data.index, self.saved)

        with open(path, "a") as data_file:
            for line in unsaved:
                line_data = ",".join(map(str, list(self.data.loc[line])))
                data_file.write(str(line) + "," + line_data + "\n")

        self.saved = self.saved + list(unsaved)

        self.status = base_status

    def start(self):
        """
        Start the experiment: clock starts/resumes, routines resume,
        measurements continue

        :return: None
        """
        self.clock.start()

        logger.info("Starting experiment")

        self.status_locked = False
        self.status = Experiment.RUNNING
        self.status_locked = True

    def hold(self, reason=None):
        """
        Hold the experiment: clock stops, routines stop, measurements continue

        :return: None
        """
        self.clock.stop()

        logger.info(f"Holding experiment ({reason})")

        self.status_locked = False
        self.status = Experiment.HOLDING
        if reason:
            self.status = self.status + ": " + reason
        self.status_locked = True

    def stop(self, reason=None):
        """
        Stop the experiment: clock stops, routines stop, measurements stop

        :return: None
        """
        self.clock.stop()

        logger.info(f"Stopping experiment ({reason})")

        self.status_locked = False
        self.status = Experiment.STOPPED
        if reason:
            self.status = self.status + ": " + reason
        self.status_locked = True

    def terminate(self, reason=None):
        """
        Terminate the experiment: clock, routines and measurements stop,
        data is saved and StopIteration is raised

        :return: None
        """

        logger.info(f"Terminating experiment ({reason})")

        self.stop()
        self.save()

        self.status_locked = False
        self.status = Experiment.TERMINATED
        if reason:
            self.status = self.status + ": " + reason
        self.status_locked = True

        # End routines
        for routine in self.routines.values():
            routine.terminate()

    def __repr__(self):
        return "Experiment"


class AsyncExperiment(Experiment):
    """
    Asynchronous version of Experiment

    Each variable and routine is updated as quickly as possible independent of the
    experiment iteration. Every time a variable is updated, the corresponding entry in
    `state` is also updated.
    """

    def __init__(
            self,
            variables: dict,
            routines: dict = None,
            end: Union[numbers.Number, str, None] = None,
    ):
        super().__init__(variables, routines, end)

    def __next__(self):

        # Start the clock and loop on first call
        if self.state.name is None:  # first step of the experiment
            self.start()

            self.loop = asyncio.get_running_loop()

            for name in self.variables:
                self.loop.create_task(self._update_variable(name))

            for name in self.routines:
                self.loop.create_task(self._update_routine(name))

            self.status = Experiment.RUNNING

        logger.info(f'Iterating experiment (t = {self.state["Time"]} s)')

        # End the experiment, if the duration of the experiment has passed
        if self.clock.time > self.end:
            self.terminate()

        if (self.running or self.holding) and self.state.name is not None:
            # Append new state to experiment data set
            with warnings.catch_warnings():
                warnings.simplefilter(action="ignore", category=FutureWarning)
                self.data.loc[self.state.name] = self.state

        elif self.terminated:
            raise StopIteration

        return self.state

    async def _update_variable(self, name):
        """Update named variable"""
        while not self.terminated:
            if self.running or self.holding:

                await asyncio.to_thread(Experiment._update_variable, self, name)

                # Update time
                self.state["Time"] = self.clock.time
                self.state.name = datetime.datetime.now()

            else:
                await asyncio.sleep(0.1)  # give other updating tasks a chance to run

    async def _update_routine(self, name):
        """Update named routine"""
        while not self.terminated:
            if self.running:

                # Update time
                self.state["Time"] = self.clock.time
                self.state.name = datetime.datetime.now()

                await asyncio.to_thread(Experiment._update_routine, self, name)

            else:
                await asyncio.sleep(0.1)  # give other updating tasks a chance to run


class Alarm:
    """
    Triggers if a condition among variables is met and indicates the response
    protocol
    """

    def __init__(self, condition, definitions, protocol=None):
        self.trigger_variable = _variables.Expression(
            expression=condition, definitions=definitions
        )

        if protocol:
            self.protocol = protocol
        else:
            self.protocol = "none"

    @property
    def triggered(self):
        return bool(self.trigger_variable.value)

    def __repr__(self):
        return "Alarm"


class Manager:
    """
    Utility class which sets up and manages experiments, based on runcards.
    When initallized, it uses the given runcard to construct an experiment and
    run it in a separate thread. The Manager also handles alarms as specified
    by the runcard.
    """

    def __init__(self, runcard=None):
        """
        :param runcard: (dict/str) runcard as a dictionary or path pointing to
        a YAML file
        """

        logger.info(f"Initializing experiment manager...")

        if runcard:
            self.runcard = runcard
        else:
            # Have user locate runcard, if none given
            root = tk.Tk()
            root.withdraw()
            self.runcard = askopenfilename(
                parent=root,
                title="Select Runcard",
                filetypes=[("YAML files", "*.yaml")],
            )

            if self.runcard == "":
                raise ValueError("a valid runcard was not selected!")

        if isinstance(self.runcard, str):

            logger.info(f"Parsing runcard from file ({self.runcard})...")

            if os.path.exists(self.runcard):
                dirname = os.path.dirname(self.runcard)
                if dirname != "":
                    os.chdir(os.path.dirname(self.runcard))
                    # go to runcard directory to put data in same location

                yaml = YAML()
                with open(self.runcard, "rb") as runcard_file:
                    self.runcard = yaml.load(runcard_file)  # load the runcard
            else:
                raise FileNotFoundError(f'invalid runcard path "{self.runcard}"')
        elif isinstance(self.runcard, dict):

            logger.info(f"Parsing runcard from dictionary...")

        else:
            raise TypeError(
                f"runcard given to Manager must be a path to a YAML file "
                f"or a dictionary, not {type(self.runcard)}!"
            )

        converted_runcard = convert_runcard(self.runcard)

        self.description = converted_runcard["Description"]
        self.settings = converted_runcard["Settings"]
        self.instruments = converted_runcard["Instruments"]
        self.alarms = converted_runcard["Alarms"]
        self.plotter = converted_runcard.get("Plotter", None)

        logger.info("Building experiment from runcard...")

        self.experiment = converted_runcard["Experiment"]

        # Unpack settings
        self.followup = self.settings.get("follow-up", None)
        self.step_interval = convert_time(self.settings.get("step interval", 0.1))
        self.save_interval = convert_time(self.settings.get("save interval", 60))
        self.plot_interval = convert_time(self.settings.get("plot interval", 0.1))

        self.last_save = 0

        self.awaiting_alarms = {}  # dictionary of alarms that are triggered

    def run(self, directory=None):
        """
        Run the experiment defined by the runcard. A GUI shows experiment
        status, while the experiment is run in a separate thread.

        :param directory: (path) (optional) directory in which to run the
                          experiment if different from the working directory
        :return: None
        """

        top_dir = os.getcwd()

        if directory:
            os.chdir(directory)

        # Create a new directory for data storage
        experiment_name = self.runcard["Description"].get("name", "Experiment")
        working_dir = experiment_name + "-" + self.experiment.timestamp
        os.mkdir(working_dir)
        os.chdir(working_dir)

        # Save executed runcard alongside data for record keeping
        yaml = YAML()

        timestamped_path = f"{experiment_name}_{self.experiment.timestamp}.yaml"
        with open(timestamped_path, "w") as runcard_file:
            yaml.dump(self.runcard, runcard_file)

        # Run experiment loop in separate thread
        logger.info(f"Starting {experiment_name}")
        experiment_thread = threading.Thread(target=asyncio.run, args=(self._run(),))
        experiment_thread.start()

        # Set up the GUI for user interaction
        logger.info("Launching GUI")
        self.gui = _graphics.ExperimentGUI(
            self.experiment,
            alarms=self.alarms,
            instruments=self.instruments,
            title=self.description.get("name", "Experiment"),
            plotter=self.plotter,
            save_interval=self.save_interval,
            plot_interval=self.plot_interval,
        )

        self.gui.run()

        experiment_thread.join()
        logger.info("Experiment complete; cleaning up...")

        # Disconnect instruments
        logger.info("Disconnecting from instruments")
        for instrument in self.instruments.values():
            instrument.disconnect()

        os.chdir(top_dir)  # return to the parent directory

        # Cancel follow-up if experiment terminated by user
        if self.experiment.terminated and "user" in self.experiment.status:
            logger.info("Cancelling follow-up experiment")
            self.followup = None

        # Execute the follow-up experiment if there is one
        if self.followup:
            logger.info("Running follow-up experiment")
            if "yaml" in self.followup:
                self.__init__(self.followup)
                self.run(directory=directory)
            elif "repeat" in self.followup:
                self.__init__(self.runcard)
                self.run(directory=directory)

    async def _run(self):
        # Run in a separate thread

        for state in self.experiment:

            logger.info(
                "state ="
                + ".,".join([f"{key}: {value}" for key, value in state.items()])
            )

            step_start = time.time()

            # Save experimental data periodically
            next_save = self.last_save + self.save_interval
            if self.experiment.clock.time >= next_save:
                save_thread = threading.Thread(target=self.experiment.save)
                save_thread.start()
                self.last_save = self.experiment.clock.time

            # Check if any alarms are triggered and handle them
            for name, alarm in self.alarms.items():
                if alarm.triggered:
                    # Add to dictionary of triggered alarms, if newly triggered
                    if name not in self.awaiting_alarms:
                        self.awaiting_alarms[name] = {
                            "time": self.experiment.data["Time"].iloc[-1],
                            "status": self.experiment.status,
                        }

                    if "none" in alarm.protocol:
                        # do nothing (GUI will indicate that alarm is triggered)
                        break
                    if "hold" in alarm.protocol:
                        # stop routines but keep measuring until alarm is clear
                        self.experiment.hold(reason=name)
                        break
                    elif "stop" in alarm.protocol:
                        # stop routines and measurements until alarm is clear
                        self.experiment.stop(reason=name)
                        break
                    elif "terminate" in alarm.protocol:
                        # terminate the experiment
                        self.experiment.terminate(reason=name)
                        break
                    elif "yaml" in alarm.protocol:
                        # terminate the experiment and run another
                        self.experiment.terminate(reason=name)
                        self.followup = alarm.protocol
                        break
                    elif "check" in alarm.protocol:
                        # stop routines and measurements until user resumes
                        self.experiment.stop(reason=name)
                        break
                    else:
                        # terminate the experiment
                        self.experiment.terminate(reason=name)
                        break

                elif name in self.awaiting_alarms:
                    # Alarm was previously triggered but is now clear

                    if "check" not in alarm.protocol:
                        # alarm is not waiting for user to check

                        info = self.awaiting_alarms.pop(name)
                        # remove from dictionary of triggered alarms
                        prior_status = info["status"]

                        if len(self.awaiting_alarms) == 0:
                            # there are no more triggered alarms
                            # return to state of experiment prior to trigger
                            if "Running" in prior_status:
                                self.experiment.start()
                            if "Holding" in prior_status:
                                self.experiment.hold(reason=name + " cleared")
                            if "Stopped" in prior_status:
                                self.experiment.stop(reason=name + " cleared")

            step_end = time.time()

            remaining_time = self.step_interval - (step_end - step_start)

            await asyncio.sleep(max([remaining_time, 0.0]))

    def __repr__(self):
        return "Manager"


class RuncardError(BaseException):
    pass


def validate_runcard(runcard):
    logger.info("Validating runcard")

    is_dict = isinstance(runcard, dict)
    is_ordereddict = isinstance(runcard, collections.OrderedDict)

    if is_dict or is_ordereddict:
        # Create temporary runcard YAML file for PyKwalify to validate
        yaml = YAML()

        runcard_path = f"tmp_runcard_{time.time()}.yaml"

        with open(runcard_path, "w") as runcard_file:
            yaml.dump(runcard, runcard_file)

        try:
            validate_runcard(runcard_path)
        finally:
            # Wait until temporary file has write access, then delete
            while not os.access(runcard_path, os.W_OK):
                time.sleep(0.1)

            os.remove(runcard_path)

        return True

    elif type(runcard) is not str:
        raise ValueError("runcard must be either dict or str.")

    validator = YamlValidator(
        source_file=runcard,
        schema_files=[
            os.path.join(pathlib.Path(__file__).parent, "runcard_schema.yaml")
        ],
    )

    try:
        validator.validate(raise_exception=True)
    except pykwalify.errors.SchemaError as err:
        raise RuncardError(err)

    return True


def convert_runcard(runcard):
    """
    :param runcard: (dict/str) runcard dictionary or path string

    :return: (dict) converted runcard in the form described below.

    Converts the sections into the relevant objects:

    * The Descriptions and Settings sections are unchanged.
    * The Instruments section is converted into a dictionary of corresponding
      ``Instrument`` instances.
    * The Variables and Routines sections are combined into an ``Experiment``
      instance.
    * The Alarms section is  converted into a dictionary of corresponding
      ``Alarm`` objects.
    * The Plots section is converted into a corresponding ``Plotter`` instance.
    """

    logger.info("Building experiment from runcard")

    # if runcard argument is a path to a YAML file, load into a dictionary
    if type(runcard) == str:
        yaml = YAML()
        with open(runcard, "rb") as runcard_file:
            runcard = yaml.load(runcard_file)

    # Validate runcard format and contents
    validate_runcard(runcard)

    converted_runcard = runcard.copy()

    # Load any custom components
    custom_routines = {}
    custom_instruments = {}

    if "custom.py" in os.listdir():
        sys.path.insert(1, os.getcwd())
        custom = importlib.import_module("custom")

        for name, thing in custom.__dict__.items():
            if type(thing) == type:
                if issubclass(thing, _routines.Routine):
                    custom_routines[name] = thing
                if issubclass(thing, _instruments.Instrument):
                    custom_instruments[name] = thing

    # Instruments section
    available_instruments = {**_instruments.supported, **custom_instruments}

    instruments = {}
    for name, specs in runcard.get("Instruments", {}).items():

        logger.info(f"Initializing instrument {name}")

        specs = specs.copy()
        _type = specs.pop("type")
        address = specs.get("address", None)

        # Grab any keyword arguments for the adapter
        adapter_kwargs = {}
        for kwarg in _adapters.kwargs:
            if kwarg.replace("_", " ") in specs:
                adapter_kwargs[kwarg] = specs.pop(kwarg.replace("_", " "))

        # Any remaining keywords are instrument presets
        presets = {
            key: recast(value) for key, value in specs.get("presets", {}).items()
        }
        postsets = {
            key: recast(value) for key, value in specs.get("postsets", {}).items()
        }

        instrument_class = available_instruments[_type]
        instruments[name] = instrument_class(
            address=address, presets=presets, postsets=postsets, **adapter_kwargs
        )
        instruments[name].name = name

    converted_runcard["Instruments"] = instruments

    # Variables section
    variables = {}
    for name, specs in runcard["Variables"].items():

        logger.info(f"Initializing variable {name}")

        if "meter" in specs:
            instrument = converted_runcard["Instruments"][specs["instrument"]]
            gate = specs.get("gate", None)
            gate = variables[gate] if gate else None
            variables[name] = _variables.Meter(
                meter=specs["meter"],
                instrument=instrument,
                gate=gate,
                multiplier=specs.get("multiplier", 1),
                offset=specs.get("offset", 0),
            )
        elif "knob" in specs:
            instrument = converted_runcard["Instruments"][specs["instrument"]]
            variables[name] = _variables.Knob(
                knob=specs["knob"],
                instrument=instrument,
                lower_limit=specs.get("lower limit", None),
                upper_limit=specs.get("upper limit", None),
                multiplier=specs.get("multiplier", 1),
                offset=specs.get("offset", 0),
            )
        elif "expression" in specs:
            expression = specs["expression"]
            definitions = specs.get("definitions", {}).copy()

            for variable in definitions.values():
                if variable not in variables:
                    raise KeyError(
                        f"variable {variable} specified for expression {name} "
                        f"is not in Variables!"
                    )

            definitions = {
                symbol: variables[name] for symbol, name in definitions.items()
            }

            variables[name] = _variables.Expression(
                expression=expression, definitions=definitions
            )
        elif "server" in specs:
            server = specs["server"]
            alias = specs.get("alias", name)
            protocol = specs.get("protocol", None)
            settable = specs.get("settable", False)

            variables[name] = _variables.Remote(
                server=server,
                alias=alias,
                protocol=protocol,
                settable=settable,
                multiplier=specs.get("multiplier", 1),
                offset=specs.get("offset", 0),
            )

        elif "parameter" in specs:
            variables[name] = _variables.Parameter(parameter=recast(specs["parameter"]))

        if name in variables and "hidden" in specs:
            # If hidden, the variable does not appear in the GUI
            if specs["hidden"]:
                variables[name]._hidden = True

    # Routines section
    available_routines = {**_routines.supported, **custom_routines}

    routines = {}
    if "Routines" in runcard:
        for name, specs in runcard["Routines"].items():

            logger.info(f"Initializing routine {name}")

            specs = specs.copy()  # avoids modifying the runcard
            _type = specs.pop("type")

            specs = {key.replace(" ", "_"): value for key, value in specs.items()}

            # Get knobs
            knobs = specs.pop("knobs", None)

            if np.isscalar(knobs):
                knobs = [knobs]
            elif isinstance(knobs, dict):  # shortcut for ModbusServer
                knob_addresses = np.array(list(knobs.keys())).flatten()
                knobs = np.array(list(knobs.values())).flatten()
                specs["knob_addresses"] = knob_addresses
            elif knobs is not None and not isinstance(knobs, list):
                raise ValueError(
                    f"Invalid knobs specification for routine {name}; "
                    "Must be a list of knob names or optionally a dictionary "
                    "of register addresses and knob names for a ModbusServer"
                )

            # Convert list of knobs into dictionary of {name: knob} pairs
            if knobs is not None and len(knobs) > 0:
                for knob in knobs:
                    if knob not in variables:
                        raise KeyError(
                            f"knob {knob} specified for routine {name} "
                            "is not in Variables!"
                        )

                specs["knobs"] = {name: variables[name] for name in knobs}
            else:
                specs["knobs"] = None

            # Get meters
            meters = specs.pop("meters", None)

            if np.isscalar(meters):
                meters = [meters]
            elif isinstance(meters, dict):  # shortcut for ModbusServer
                meter_addresses = np.array(list(meters.keys())).flatten()
                meters = np.array(list(meters.values())).flatten()

                specs["meter_addresses"] = meter_addresses
            elif meters is not None and not isinstance(meters, list):
                raise ValueError(
                    f"Invalid meters specification for routine {name}; "
                    "Must be a list of meter names or optionally a dictionary "
                    "of register addresses and meter names for a ModbusServer"
                )

            specs["meters"] = meters

            routines[name] = available_routines[_type](**specs)

    # Create the experiment
    async_experiment = False
    if "Settings" in runcard:
        if runcard["Settings"].get("async", False):
            async_experiment = True

    if async_experiment:

        logger.info("Initializing synchronous experiment")

        converted_runcard["Experiment"] = AsyncExperiment(
            variables, routines=routines, end=runcard["Settings"].get("end", None)
        )
    else:

        logger.info("Initializing asynchronous experiment")

        converted_runcard["Experiment"] = Experiment(
            variables, routines=routines, end=runcard["Settings"].get("end", None)
        )

    # Alarms section
    alarms = {}
    if "Alarms" in runcard:
        for name, specs in runcard["Alarms"].items():

            logger.info(f"Initializing alarm {name}")

            condition = specs.copy()["condition"]
            definitions = specs.copy().get("definitions", {})

            for variable in definitions.values():
                if variable not in variables:
                    raise KeyError(
                        f"variable {variable} specified for alarm {name} "
                        f"is not in Variables!"
                    )

            definitions = {
                symbol: variables[name] for symbol, name in definitions.items()
            }

            alarms.update(
                {
                    name: Alarm(
                        condition, definitions, protocol=specs.get("protocol", None)
                    )
                }
            )

    converted_runcard["Alarms"] = alarms

    # Plots section
    if "Plots" in runcard:

        logger.info("Initializing plotter")

        converted_runcard["Plotter"] = _graphics.Plotter(
            converted_runcard["Experiment"].data, settings=runcard["Plots"]
        )
    else:
        converted_runcard["Plotter"] = None

    return converted_runcard



import re
import struct
import socket
import numpy as np

from empyric.tools import logger
from empyric.types import Toggle, Float, ON, OFF
from empyric.adapters import Serial, Socket, GPIB
from empyric.collection.instrument import Instrument, setter, getter, measurer


class Keithley2260B(Instrument):
    """
    Keithley 2260B power supply, usually either 360 W or 720 W
    """

    name = "Keithley2260B"

    supported_adapters = (
        (Serial, {"baud_rate": 115200, "write_termination": "\r\n"}),
        (Socket, {"read_termination": "\n", "write_termination": "\n"}),
    )

    knobs = ("max voltage", "max current", "output")

    # no presets or postsets for this instrument

    meters = ("voltage", "current")

    @measurer
    def measure_current(self) -> Float:
        def validator(response):
            return bool(re.match("[\+\-]\d+\.\d\d\d", response.strip()))

        return float(self.query("MEAS:CURR?", validator=validator))

    @measurer
    def measure_voltage(self) -> Float:
        def validator(response):
            return bool(re.match("[\+\-]\d+\.\d\d\d", response.strip()))

        return float(self.query("MEAS:VOLT?", validator=validator))

    @setter
    def set_max_voltage(self, voltage: Float):
        self.write("VOLT %.4f" % voltage)

    @setter
    def set_max_current(self, current: Float):
        self.write("CURR %.4f" % current)

    @setter
    def set_output(self, output: Toggle):
        if output == ON:
            self.write("OUTP:STAT:IMM ON")
        elif output == OFF:
            self.write("OUTP:STAT:IMM OFF")

    @getter
    def get_output(self) -> Toggle:
        response = self.query("OUTP?")

        if response == "0":
            return OFF
        elif response == "1":
            return ON

    @getter
    def get_max_current(self) -> Float:
        def validator(response):
            return bool(re.match("[\+\-]\d+\.\d\d\d", response.strip()))
        
        return float(self.query("CURR?", validator=validator))

    @getter
    def get_max_voltage(self) -> Float:
        def validator(response):
            return bool(re.match("[\+\-]\d+\.\d\d\d", response.strip()))

        return float(self.query("VOLT?", validator=validator))


class BK9183B(Instrument):
    """
    B&K Precision Model 9183B (35V & 6A / 70V & 3A) power supply
    """

    name = "BK9183B"

    supported_adapters = (
        (Serial, {"baud_rate": 57600}),
        (Socket, {"read_termination": "\n", "write_termination": "\n"}),
    )

    knobs = ("max voltage", "max current", "output")

    # no presets or postsets for this instrument

    meters = ("voltage", "current")

    @setter
    def set_output(self, output: Toggle):
        if output == ON:
            self.write("OUT ON")
        elif output == OFF:
            self.write("OUT OFF")

    @measurer
    def measure_current(self):
        # sometimes the first measurement is lagged
        return [float(self.query("MEAS:CURR?")) for i in range(3)][-1]

    @measurer
    def measure_voltage(self):
        # sometimes the first measurement is lagged
        return [float(self.query("MEAS:VOLT?")) for i in range(3)][-1]

    @setter
    def set_max_current(self, current):
        self.write("SOUR:CURR " + str(current))

    @setter
    def set_max_voltage(self, voltage):
        self.write("SOUR:VOLT " + str(voltage))

    @getter
    def get_max_current(self):
        return float(self.query("SOUR:CURR?"))

    @getter
    def get_max_voltage(self):
        return float(self.query("SOUR:VOLT?"))


class UltraflexInductionHeater(Instrument):
    """UltraFlex Ultraheat S2 (or similar) 2 kW induction heater"""

    name = "UltraflexInductionHeater"

    supported_adapters = (Serial, {})

    knobs = ("power", "output", "mode")

    meters = (
        "voltage",
        "current",
        "frequency",
    )

    @measurer
    def measure_current(self):
        return 0.1 * float(self.query("S3i11K")[3:-3], 16)

    @measurer
    def measure_voltage(self):
        return float(self.query("S3v04K")[3:-3], 16)

    @measurer
    def measure_frequency(self):
        return 1e3 * float(self.query("S3f14K")[3:-3], 16)

    @setter
    def set_power(self, percent_of_max):
        if percent_of_max < 0 or percent_of_max > 100:
            raise ValueError(
                "power setting must be a percentage value between 0 and 100"
            )

        ascii_command = "S34"
        ascii_command += hex(int(round(percent_of_max)))[-2:].upper()
        ascii_command += hex(512 - sum(ascii_command.encode()))[-2:].upper()
        self.query(ascii_command + "K")

    @getter
    def get_power(self):
        return float(self.query("S3B38K")[3:5], 16)

    @setter
    def set_output(self, output):
        if Toggle(output):
            self.query("S3200E8K")
        else:
            self.query("S3347K")

    @setter
    def set_mode(self, mode):
        if mode.lower() == "manual":
            self.query("S3L01CDK")
        elif mode.lower() == "remote":
            self.query("S3L00CEK")
        else:
            raise ValueError(
                f"{self.name}: Unsupported control mode {mode}. "
                'Allowed values are "manual" and "remote".'
            )


class SRSPS300(Instrument):
    """
    Stanford Rsearch Systems PS-300 series high voltage power supply.
    """

    name = "SRSPS300"

    supported_adapters = ((GPIB, {}),)

    knobs = (
        "output",
        "voltage",
        "max voltage",
        "max current",
        "trip current",
        "clear trip",
    )

    presets = {"output": "OFF", "voltage": 0, "max current": 5e-3}

    postsets = {"output": "OFF", "voltage": 0, "max current": 5e-3}

    meters = ("voltage", "current")

    @setter
    def set_output(self, output: Toggle):
        if output == ON:
            self.write("HVON")
        else:
            self.write("HVOF")

    @getter
    def get_output(self) -> Toggle:
        # last bit of status byte is the output state

        def validator(response):
            return re.match("\d{1}", response)

        status_bit_7 = int(self.query("*STB? 7", validator=validator))

        if status_bit_7 == 1:
            return ON
        else:
            return OFF

    @setter
    def set_voltage(self, voltage: Float):
        self.write("VSET%f" % float(voltage))

    @getter
    def get_voltage(self) -> Float:
        return float(self.query("VSET?"))

    @setter
    def set_max_voltage(self, voltage: Float):
        self.write("VLIM%f" % float(voltage))

    @getter
    def get_max_voltage(self) -> Float:
        return float(self.query("VLIM?"))

    @setter
    def set_max_current(self, current: Float):
        self.write("ILIM%f" % float(current))  # all units are SI units-->Amps

    @getter
    def get_max_current(self) -> Float:
        return float(self.query("ILIM?"))

    @setter
    def set_trip_current(self, current: Float):
        self.write("ITRP%f" % float(current))

    @getter
    def get_trip_current(self) -> Float:
        return float(self.query("ITRP?"))

    @setter
    def set_clear_trip(self, state: Toggle) -> Toggle:
        if state == ON:
            self.write("TCLR")

        return self.get_clear_trip()

    @getter
    def get_clear_trip(self):

        try:
            status_byte_1 = int(self.query("*STB? 1"))
            status_byte_2 = int(self.query("*STB? 2"))

            ovp_tripped = status_byte_1 == 1
            ocp_tripped = status_byte_2 == 1

        except ValueError:
            return None

        if ovp_tripped or ocp_tripped:
            return OFF
        else:
            return ON

    @measurer
    def measure_voltage(self) -> Float:
        return float(self.query("VOUT?"))

    @measurer
    def measure_current(self) -> Float:
        return float(self.query("IOUT?"))


class KoradKWR100(Instrument):
    """
    Korad KWR100 series power supply

    When using LAN communications (via Socket adapter) use the KWR100 Assistant
    executable to configure IP address and port number. Also, send the command
    ":SYST:UPDMODE 0" in order to be able to use an arbitrary local port.
    """

    name = "KWR100"

    supported_adapters = (
        (
            Socket,
            {
                "type": socket.SOCK_DGRAM,  # UDP
                "write_termination": "\n",
                "read_termination": "\n",
                "timeout": 1,
            },
        ),
    )

    knobs = ("output", "max voltage", "max current")

    meters = ("voltage", "current")

    @setter
    def set_output(self, state: Toggle):
        if state == ON:
            self.write("OUT:1")
        elif state == OFF:
            self.write("OUT:0")

    @getter
    def get_output(self) -> Toggle:
        response = self.query("OUT?")

        if response == "0":
            return OFF
        elif response == "1":
            return ON

    @setter
    def set_max_voltage(self, voltage: Float):
        self.write("VSET:%.2f" % voltage)

    @getter
    def get_max_voltage(self) -> Float:
        return float(self.query("VSET?"))

    @setter
    def set_max_current(self, current: Float):
        self.write("ISET:%.2f" % current)

    @getter
    def get_max_current(self) -> Float:
        return float(self.query("ISET?"))

    @measurer
    def measure_voltage(self) -> Float:
        return float(self.query("VOUT?"))

    @measurer
    def measure_current(self) -> Float:
        return float(self.query("IOUT?"))


class MagnaPowerSL1000(Instrument):
    """Magna-Power SL series 1.5 kW (1 kV / 1.5 A) power supply"""

    supported_adapters = (
        (
            Serial,
            {
                "baud_rate": 19200,
                "read_termination": "\r\n",
                "write_termination": "\r\n",
                "timeout": 10.0,
            },
        ),
    )

    name = "MagnaPowerSL1000"

    knobs = (
        "output",
        "output protection clear",
        "over voltage protection",
        "over current protection",
        "max voltage",
        "max current",
    )

    meters = ("voltage", "current")

    @setter
    def set_output(self, state: Toggle):
        if state == ON:
            self.write("OUTP:START")
        elif state == OFF:
            self.write("OUTP:STOP")

    @getter
    def get_output(self) -> Toggle:
        response = self.query("OUTP?").strip()

        if response == b"1":
            return ON
        elif response == b"0":
            return OFF
        else:
            return None

    @setter
    def set_output_protection_clear(self, state: Toggle):
        if state == ON:
            self.write("OUTP:PROT:CLE")

        return self.get_over_current_protection()

    @getter
    def get_output_protection_clear(self) -> Toggle:
        # Read the "questionable" register

        response = self.query("STAT:QUES:COND?")

        try:
            register = int(response)
        except ValueError:
            return None

        # Questionable Register bits as described in the manual
        ov_bit, register = register // 1024, register % 1024
        oc_bit, register = register // 512, register % 512
        pb_bit, register = register // 256, register % 256
        pgm_bit, register = register // 128, register % 128
        ot_bit, register = register // 64, register % 64
        fuse_bit, register = register // 32, register % 32
        alm_bit, register = register // 16, register % 16

        if any([ov_bit, oc_bit, pb_bit, pgm_bit, ot_bit, fuse_bit, alm_bit]):
            return OFF
        else:
            return ON

    @setter
    def set_over_voltage_protection(self, voltage: Float):
        self.write("VOLT:PROT %.1f" % voltage)

    @getter
    def get_over_voltage_protection(self) -> Float:
        response = self.query("VOLT:PROT?")

        try:
            return float(response)
        except ValueError:
            return np.nan

    @setter
    def set_over_current_protection(self, voltage: Float):
        self.write("CURR:PROT %.1f" % voltage)

    @getter
    def get_over_current_protection(self) -> Float:
        response = self.query("CURR:PROT?")

        try:
            return float(response)
        except ValueError:
            return np.nan

    @setter
    def set_max_voltage(self, voltage: Float):
        self.write("VOLT %.1f" % voltage)

    @getter
    def get_max_voltage(self) -> Float:
        response = self.query("VOLT?")

        try:
            return float(response)
        except ValueError:
            return np.nan

    @setter
    def set_max_current(self, current: Float):
        self.write("CURR %.1f" % current)

    @getter
    def get_max_current(self) -> Float:
        response = self.query("CURR?")

        try:
            return float(response)
        except ValueError:
            return np.nan

    @measurer
    def measure_voltage(self) -> Float:
        response = self.query("MEAS:VOLT?").strip()

        try:
            return float(response)
        except ValueError:
            return np.nan

    @measurer
    def measure_current(self) -> Float:
        response = self.query("MEAS:CURR?").strip()

        try:
            return float(response)
        except ValueError:
            return np.nan


class SorensenXG10250(Instrument):
    """
    Sorensen XG 10-250 series high current power supply.

    The analog control mode option allows the power supply to operate in a
    voltage-controlled current mode via its non-isolated input pin.

    By default, the RS-485 multicast address is assumed to be 1.
    """

    name = "SorensenXG10250"

    supported_adapters = (
        (Serial, {"baud_rate": 9600, "read_termination": "\r", "lib": "pyserial"}),
    )

    knobs = ("max voltage", "max current", "output", "analog control mode")

    meters = ("voltage", "current", "analog input voltage", "analog input current")

    def __init__(
        self, address=None, adapter=None, presets=None, postsets=None, **kwargs
    ):
        self.address = address

        self.knobs = ("connected",) + self.knobs

        self.analog_mode_state = None

        adapter_connected = False
        if adapter:
            self.adapter = adapter(self, **kwargs)
        else:
            errors = []
            for _adapter, settings in self.supported_adapters:
                settings.update(kwargs)
                try:
                    self.adapter = _adapter(self, **settings)
                    adapter_connected = True
                    break
                except BaseException as error:
                    msg = (
                        f"in trying {_adapter.__name__} adapter, "
                        f"got {type(error).__name__}: {error}"
                    )
                    errors.append(msg)

            if not adapter_connected:
                message = (
                    f"unable to connect an adapter to "
                    f"instrument {self.name} at address {address}:\n"
                )
                for error in errors:
                    message = message + f"{error}\n"
                raise ConnectionError(message)

        if self.address:
            self.name = self.name + "@" + str(self.address)

        self.write("*ADR 1")  # Set multicast address to 1 by default

        # Get existing knob settings, if possible
        for knob in self.knobs:
            if hasattr(self, "get_" + knob.replace(" ", "_")):
                # retrieves the knob value from the instrument
                self.__getattribute__("get_" + knob.replace(" ", "_"))()
            else:
                # knob value is unknown until it is set
                self.__setattr__(knob.replace(" ", "_"), None)

        # Apply presets
        if presets:
            self.presets = {**self.presets, **presets}

        for knob, value in self.presets.items():
            self.set(knob, value)

        # Get postsets
        if postsets:
            self.postsets = {**self.postsets, **postsets}

        self.kwargs = kwargs

    def float_validator(self, response):
        return bool(re.match("\d+\.\d+", response))

    @measurer
    def measure_current(self):
        return self.query("MEAS:CURR?", validator=self.float_validator)

    @measurer
    def measure_voltage(self):
        return self.query("MEAS:VOLT?", validator=self.float_validator)

    @measurer
    def measure_analog_input_voltage(self):
        return self.query("MEAS:APR?", validator=self.float_validator)

    @measurer
    def measure_analog_input_current(self):
        return self.query("MEAS:APR:CURR?", validator=self.float_validator)

    @setter
    def set_output(self, output: Toggle):
        if output == ON:
            self.write("OUTP ON")
        elif output == OFF:
            self.write("OUTP OFF")

    @setter
    def set_analog_control_mode(self, analog_control_mode: Toggle):
        if analog_control_mode == ON:
            self.write("SYST:REM:SOUR:CURR AVOL")
        if analog_control_mode == OFF:
            self.write("SYST:REM:SOUR:CURR LOC")

        # Verify mode was sent
        def str_validator(response):
            return bool(re.match("Current mode: .*", response))

        response = self.query(
            "SYST:REM:SOUR:CURR?", validator=str_validator, until="\r"
        )
        if "Analog " in response:
            self.analog_mode_state = ON
        elif "LOCAL" in response:
            self.analog_mode_state = OFF

    @setter
    def set_max_current(self, current):
        self.write("SOUR:CURR " + str(current))

    @setter
    def set_max_voltage(self, voltage):
        self.write("SOUR:VOLT " + str(voltage))

    @getter
    def get_max_current(self):
        return self.query("SOUR:CURR?")

    @getter
    def get_max_voltage(self):
        return self.query("SOUR:VOLT?")

    @getter
    def get_output(self) -> Toggle:
        response = self.query("OUTP?")
        if response.startswith("0"):
            return OFF
        elif response.startswith("1"):
            return ON

    @getter
    def get_analog_control_mode(self) -> Toggle:
        if self.analog_mode_state is None:

            def str_validator(response):
                return bool(re.match("Current mode: .*", response))

            response = self.query(
                "SYST:REM:SOUR:CURR?", validator=str_validator, until="\r"  # :CURR?",
            )
            if "Analog " in response:
                self.analog_mode_state = ON
            elif "LOCAL" in response:
                self.analog_mode_state = OFF

        return self.analog_mode_state


class BK9140(Instrument):
    """
    B&K Precision Model 9140 (35V & 6A / 70V & 3A) power supply
    """

    name = "BK9140"

    supported_adapters = (
        (Serial, {"baud_rate": 57600}),
        (Socket, {"read_termination": "\n", "write_termination": "\n"}),
    )

    knobs = (
        "max voltage 1",
        "max current 1",
        "output 1",
        "max voltage 2",
        "max current 2",
        "output 2",
        "max voltage 3",
        "max current 3",
        "output 3",
    )

    # no presets or postsets for this instrument

    meters = (
        "voltage 1",
        "current 1",
        "voltage 2",
        "current 2",
        "voltage 3",
        "current 3",
    )

    @setter
    def set_output_1(self, output: Toggle):
        self.write(":INST:SEL 1")
        if output == ON:
            self.write("OUT ON")
        elif output == OFF:
            self.write("OUT OFF")

    @setter
    def set_output_2(self, output: Toggle):
        self.write(":INST:SEL 2")
        if output == ON:
            self.write("OUT ON")
        elif output == OFF:
            self.write("OUT OFF")

    @setter
    def set_output_3(self, output: Toggle):
        self.write(":INST:SEL 3")
        if output == ON:
            self.write("OUT ON")
        elif output == OFF:
            self.write("OUT OFF")

    @measurer
    def measure_current_1(self):
        # sometimes the first measurement is lagged
        self.write(":INST:SEL 1")
        return [float(self.query("MEAS:CURR?")) for i in range(3)][-1]

    @measurer
    def measure_voltage_1(self):
        # sometimes the first measurement is lagged
        self.write(":INST:SEL 1")
        return [float(self.query("MEAS:VOLT?")) for i in range(3)][-1]

    @setter
    def set_max_current_1(self, current):
        self.write(":INST:SEL 1")
        self.write("SOUR:CURR " + str(current))

    @setter
    def set_max_voltage_1(self, voltage):
        self.write(":INST:SEL 1")
        self.write("SOUR:VOLT " + str(voltage))

    @getter
    def get_max_current_1(self):
        self.write(":INST:SEL 1")
        return float(self.query("SOUR:CURR?"))

    @getter
    def get_max_voltage_1(self):
        self.write(":INST:SEL 1")
        return float(self.query("SOUR:VOLT?"))

    @measurer
    def measure_current_2(self):
        # sometimes the first measurement is lagged
        self.write(":INST:SEL 2")
        return [float(self.query("MEAS:CURR?")) for i in range(3)][-1]

    @measurer
    def measure_voltage_2(self):
        # sometimes the first measurement is lagged
        self.write(":INST:SEL 2")
        return [float(self.query("MEAS:VOLT?")) for i in range(3)][-1]

    @setter
    def set_max_current_2(self, current):
        self.write(":INST:SEL 2")
        self.write("SOUR:CURR " + str(current))

    @setter
    def set_max_voltage_2(self, voltage):
        self.write(":INST:SEL 2")
        self.write("SOUR:VOLT " + str(voltage))

    @getter
    def get_max_current_2(self):
        self.write(":INST:SEL 2")
        return float(self.query("SOUR:CURR?"))

    @getter
    def get_max_voltage_2(self):
        self.write(":INST:SEL 2")
        return float(self.query("SOUR:VOLT?"))

    @measurer
    def measure_current_3(self):
        # sometimes the first measurement is lagged
        self.write(":INST:SEL 3")
        return [float(self.query("MEAS:CURR?")) for i in range(3)][-1]

    @measurer
    def measure_voltage_3(self):
        # sometimes the first measurement is lagged
        self.write(":INST:SEL 3")
        return [float(self.query("MEAS:VOLT?")) for i in range(3)][-1]

    @setter
    def set_max_current_3(self, current):
        self.write(":INST:SEL 3")
        self.write("SOUR:CURR " + str(current))

    @setter
    def set_max_voltage_3(self, voltage):
        self.write(":INST:SEL 3")
        self.write("SOUR:VOLT " + str(voltage))

    @getter
    def get_max_current_3(self):
        self.write(":INST:SEL 3")
        return float(self.query("SOUR:CURR?"))

    @getter
    def get_max_voltage_3(self):
        self.write(":INST:SEL 3")
        return float(self.query("SOUR:VOLT?"))


class GlassmanOQ500(Instrument):
    """
    Glassman OQ series high voltage power supply (500 kV / 10 mA).


    """

    name = "GlassmanOQ"

    supported_adapters = (
        (
            Serial,
            {
                "baud_rate": 9600,
                "write_termination": "\r",
                "read_termination": "\r",
                "lib": "pyserial",
            },
        ),
    )

    knobs = ("max voltage", "max current", "output enable", "reset")

    meters = ("voltage", "current", "fault state")

    SOH: bytes = b"\x01"
    EOM: bytes = b"\x0D"

    max_output_voltage_volts = 500000.0  # 500kV
    max_output_current_mA = 20  # 20 mA

    vi_setpoints = [None, None]

    def _compute_checksum(self, message_segment: bytes) -> bytes:
        crc = (
            sum(struct.unpack(">" + "B" * len(message_segment), message_segment)) % 256
        )
        return bytes(format(crc, "X"), "utf-8")

    def _test_checksum(self, message: bytes) -> bool:
        calculated_crc = self._compute_checksum(message[1:-2])
        return message[-2:] == calculated_crc

    def _wrap_message(self, message_content) -> str:
        if type(message_content) != bytes:
            message_content = message_content.encode("utf-8")
        message = self.SOH
        message += message_content
        calculated_crc = self._compute_checksum(message_content)
        message += calculated_crc
        # message += self.EOM
        return message.decode()

    def _construct_set_message(
        self,
        normalized_voltage_cmd: float | None = None,
        normalized_current_cmd: float | None = None,
        hv_on_cmd: bool | None = None,
        hv_off_cmd: bool | None = None,
        reset_cmd: bool | None = None,
    ) -> str:
        # process inputs
        if normalized_voltage_cmd is not None:
            self.normalized_voltage_cmd = normalized_voltage_cmd
        if normalized_current_cmd is not None:
            self.normalized_current_cmd = normalized_current_cmd
        if hv_on_cmd is None:
            self.hv_on_cmd = False
        else:
            self.hv_on_cmd = hv_on_cmd
        if hv_off_cmd is None:
            self.hv_off_cmd = False
        else:
            self.hv_off_cmd = hv_off_cmd
        if reset_cmd is None:
            self.reset_cmd = False
        else:
            self.reset_cmd = reset_cmd

        # construct message
        message: bytes = b"S"  # command identifier
        message += bytes(
            format(int(0xFFF * self.normalized_voltage_cmd), "X").zfill(3), "utf-8"
        )
        message += bytes(
            format(int(0xFFF * self.normalized_current_cmd), "X").zfill(3), "utf-8"
        )
        message += b"000000"  # bytes 9-14
        message += bytes(
            format(
                int((self.reset_cmd << 2) ^ (self.hv_on_cmd << 1) ^ self.hv_off_cmd),
                "X",
            ),
            "utf-8",
        )
        return self._wrap_message(message)

    def _check_response_message(self, message) -> bool | None:
        if message == "":
            return None
        if self._test_checksum(message.encode("utf-8")):
            if message[0] == "R":
                # Check for fault state on byte 11, bit 1
                ps_fault = message[10:13][1:2]
                if ps_fault == "1":
                    logger.warning(
                        "GlassmanOQ500: Power supply is in a fault state. "
                        "A PS reset command must be sent "
                        "(via 'reset' knob) to clear fault "
                        "before setting new values!"
                    )
                return message
            else:
                raise KeyError(
                    f"GlassmanOQ500: Incorrect Message Type [{message[0]}]"
                    f" in decode_response_message()"
                )
        else:
            print(f"Checksum failed: {message}")
            raise ValueError(
                "GlassmanOQ500: Checksum error in " "decode_response_message()"
            )

    def _acknowledge_validator(self, message) -> bool | None:
        if message == "":
            return None
        if message == "A":
            return True
        elif message[0] == "E":
            logger.warning(
                f"GlassmanOQ500: Error message received "
                f"during set command: {message}. "
                f"See manual for further details."
            )
        else:
            return False

    @setter
    def set_max_voltage(self, voltage_Volts: Float):
        self.vi_setpoints[0] = voltage_Volts
        if self.vi_setpoints[1] is None:
            logger.warning(
                f"GlassmanOQ500: Waiting for current setpoint "
                f"to set voltage to {voltage_Volts} V."
            )
        else:
            normalized_voltage_cmd = voltage_Volts / self.max_output_voltage_volts
            normalized_current_cmd = self.vi_setpoints[1] / self.max_output_current_mA
            message: str = self._construct_set_message(
                normalized_voltage_cmd=normalized_voltage_cmd,
                normalized_current_cmd=normalized_current_cmd,
            )
            self.query(message, validator=self._acknowledge_validator)

    @setter
    def set_max_current(self, current_mA: Float):
        self.vi_setpoints[1] = current_mA
        if self.vi_setpoints[0] is None:
            logger.warning(
                f"GlassmanOQ500: Waiting for voltage setpoint "
                f"to set current to {current_mA} mA."
            )
        else:
            normalized_voltage_cmd = (
                self.vi_setpoints[0] / self.max_output_voltage_volts
            )
            normalized_current_cmd = current_mA / self.max_output_current_mA
            message: str = self._construct_set_message(
                normalized_voltage_cmd=normalized_voltage_cmd,
                normalized_current_cmd=normalized_current_cmd,
            )
            self.query(message, validator=self._acknowledge_validator)

    @getter
    def get_output_enable(self) -> Toggle:
        query = self._wrap_message("Q")
        response: str = self.query(query)
        response = self._check_response_message(response)[10:13][0:1]
        bytestr = bin(int(response))[2:].zfill(4)
        bit = bytestr[1:2]
        if bit == "1":
            return ON
        else:
            return OFF

    @setter
    def set_output_enable(self, output: Toggle):
        if output == ON:
            if self.vi_setpoints[0] is None or self.vi_setpoints[1] is None:
                logger.warning(
                    "GlassmanOQ500: Waiting for voltage and current "
                    "setpoints to be set in order to set output ON."
                )
            else:
                normalized_voltage_cmd = (
                    self.vi_setpoints[0] / self.max_output_voltage_volts
                )
                normalized_current_cmd = (
                    self.vi_setpoints[1] / self.max_output_current_mA
                )
                message: str = self._construct_set_message(
                    hv_on_cmd=True,
                    hv_off_cmd=False,
                    normalized_voltage_cmd=normalized_voltage_cmd,
                    normalized_current_cmd=normalized_current_cmd,
                )
                self.query(message, validator=self._acknowledge_validator)
        else:
            message: str = self._construct_set_message(
                hv_off_cmd=True,
                hv_on_cmd=False,
                normalized_voltage_cmd=0.0,
                normalized_current_cmd=0.0,
            )

            self.query(message, validator=self._acknowledge_validator)

    @setter
    def set_reset(self, reset: Toggle):
        if reset == ON:
            message: str = self._construct_set_message(
                reset_cmd=True, normalized_current_cmd=0.0, normalized_voltage_cmd=0.0
            )
            self.query(message, validator=self._acknowledge_validator)
            message: str = self._construct_set_message(
                reset_cmd=False, normalized_current_cmd=0.0, normalized_voltage_cmd=0.0
            )
            self.query(message, validator=self._acknowledge_validator)
            return OFF

    @measurer
    def measure_voltage(self) -> Float:
        response = self.query(self._wrap_message("Q"))
        voltage = self._check_response_message(response)[1:4]
        voltage_normalized = int(voltage, 16) / 0x3FF
        voltage_outp = voltage_normalized * self.max_output_voltage_volts
        return voltage_outp

    @measurer
    def measure_current(self) -> Float:
        response = self.query(self._wrap_message("Q"))
        current = self._check_response_message(response)[4:7]
        current_normalized = int(current, 16) / 0x3FF
        current_outp = current_normalized * self.max_output_current_mA
        return current_outp

    @measurer
    def measure_fault_state(self) -> Toggle:
        response = self.query(self._wrap_message("Q"))
        response = self._check_response_message(response)[10:13][1:2]
        bytestr = bin(int(response))[2:].zfill(4)
        bit = bytestr[2:3]
        if bit == "1":
            logger.warning(
                "GlassmanOQ500: Power supply is in a fault state. A PS reset "
                "command must be sent (via 'reset' knob) to clear "
                "fault before setting new values!"
            )
            return ON  # fault detected
        elif bit == "0":
            return OFF  # no fault


class PWX1500L(Instrument):
    name = "PWX1500L"

    supported_adapters = (
        (Socket, {"read_termination": "\r\n", "write_termination": "\r\n"}),
    )

    knobs = ("max voltage", "max current", "output")

    meters = ("voltage", "current")

    @measurer
    def measure_current(self) -> Float:
        def validator(response):
            return bool(re.match("[\+\-]\d+\.\d\d\d", response))

        return float(self.query("MEAS:CURR?", validator=validator))

    @measurer
    def measure_voltage(self) -> Float:
        def validator(response):
            return bool(re.match("[\+\-]\d+\.\d\d\d", response))

        return float(self.query("MEAS:VOLT?", validator=validator))

    @setter
    def set_max_voltage(self, voltage: Float):
        self.write("VOLT %.2f" % voltage)

    @setter
    def set_max_current(self, current: Float):
        self.write("CURR %.2f" % current)

    @setter
    def set_output(self, output: Toggle):
        if output == ON:
            self.write("OUTP ON")
        elif output == OFF:
            self.write("OUTP OFF")

    @getter
    def get_output(self) -> Toggle:
        response = self.query("OUTP?")

        if response == "0":
            return OFF
        elif response == "1":
            return ON

    @getter
    def get_max_current(self) -> Float:
        def validator(response):
            return bool(re.match("[\+\-]\d+\.\d\d\d", response))

        return float(self.query("CURR?", validator=validator))

    @getter
    def get_max_voltage(self) -> Float:
        def validator(response):
            return bool(re.match("[\+\-]\d+\.\d\d\d", response))

        return float(self.query("VOLT?", validator=validator))



# Experiment variables

import numbers
import socket
import time
import typing
from functools import wraps

import dill
import numpy as np  # used in Expression's eval call

from empyric.collection.instrument import Instrument

from empyric.instruments import ModbusClient
from empyric.tools import write_to_socket, read_from_socket, logger
from empyric.types import supported as supported_types, recast
from empyric.types import Type, Boolean, Float, Integer, Toggle, ON, Array


class Variable:
    """
    Base class representing a specific quantity of interest monitored in an
    experiment.
    """

    _type = None  # the data type of the variable

    #: time since the epoch of last evaluation in seconds, being equal to the
    #: result of `time.time()` being called upon the most recent evaluation of
    #: the `value` property
    last_evaluation = None

    _settable = False  # whether the variable can be set by the user
    _value = None  # last known value of the variable

    _hidden = False  # used by GUIs

    @property
    def type(self):
        """Data type of the variable"""
        return self._type

    @property
    def settable(self):
        """Whether values can be assigned to the variable by the user"""
        return self._settable

    @property
    def value(self):
        """The value of the variable"""
        # overwritten by child classes
        return

    @value.setter
    def value(self, value):
        # overwritten by child classes
        pass

    @staticmethod
    def setter_type_validator(setter):
        """Checks that set value is compatible with variable's type"""

        @wraps(setter)
        def wrapped_setter(self, value):
            if not isinstance(value, Array) and (
                value is None or value == float("nan")
            ):
                self._value = None

            elif self._type is not None:
                setter(self, recast(value, to=self._type))
            else:
                # if type is not explicitly defined upon construction,
                # infer from first set value

                recasted_value = recast(value)

                for _type in supported_types.values():
                    if isinstance(recasted_value, _type):
                        self._type = _type
                        setter(self, recasted_value)

        return wrapped_setter

    @staticmethod
    def getter_type_validator(getter):
        """Checks that get value is compatible with variable's type"""

        @wraps(getter)
        def wrapped_getter(self):
            value = getter(self)

            if not isinstance(value, Array) and (
                value is None or value == float("nan")
            ):
                self._value = None

            elif self._type is not None:
                self._value = recast(value, to=self._type)
            else:
                # if type is not explicitly defined upon construction,
                # infer from first set value

                recasted_value = recast(value)

                for _type in supported_types.values():
                    if isinstance(recasted_value, _type):
                        self._type = _type
                        self._value = recasted_value

            return self._value

        return wrapped_getter

    def __mul__(self, other):
        if isinstance(other, Variable):
            return self._value * other._value
        else:
            return self._value * other

    def __add__(self, other):
        if isinstance(other, Variable):
            return self._value + other._value
        else:
            return self._value + other

    def __sub__(self, other):
        if isinstance(other, Variable):
            return self._value - other._value
        else:
            return self._value - other

    def __bool__(self):
        return bool(self._value)

    def __eq__(self, other):
        if isinstance(other, Variable):
            return self._value == other._value
        else:
            return self._value == other


class Knob(Variable):
    """
    Variable that can be directly controlled by an instrument, such as the
    voltage of a power supply.

    The `instrument` argument specifies which instrument the knob is
    associated with.

    The `knob` argument is the label of the knob on the instrument.

    The optional `lower_limit` and `upper_limit` keyword arguments set the lower and
    upper limits for the knob. If an attempt it made to set a value above/below the
    upper/lower limit, it is set to the upper/lower limit value.

    The optional `multiplier` and `offset` keyword arguments provide a means to affect
    a linear transformation of the raw knob value. Readings from the instrument will
    be multiplied by the `multiplier` and then increased by the `offset`. Set commands
    to the instrument will take the knob value, subtract the `offset` and divide by the
    `multiplier`.
    """

    _settable = True  #:

    def __init__(
        self,
        instrument: Instrument,
        knob: str,
        lower_limit: typing.Union[float, int] = None,
        upper_limit: typing.Union[float, int] = None,
        multiplier: typing.Union[float, int] = 1,
        offset: typing.Union[float, int] = 0,
    ):
        self.instrument = instrument
        self.knob = knob  # name of the knob on instrument

        self.lower_limit = lower_limit
        self.upper_limit = upper_limit

        if self.upper_limit is not None and self.lower_limit is not None:
            if self.upper_limit < self.lower_limit:
                raise ValueError("Upper limit must be greater than lower limit")

        self.multiplier = multiplier
        self.offset = offset

        # infer type from type hint of first argument of set method
        set_method = getattr(instrument, "set_" + knob.replace(" ", "_"))
        type_hints = typing.get_type_hints(set_method)
        type_hints.pop("return", None)  # exclude return type hint

        if type_hints:
            arg_hints = list(type_hints)
            self._type = type_hints[arg_hints[0]]
        else:

            logger.warning(
                f"Unable to determine data dtype of {knob} on {instrument}; "
                "assuming 64-bit float"
            )

            self._type = np.float64

        self._value = None

    @property
    @Variable.getter_type_validator
    def value(self):
        """
        Value of the knob of an instrument
        """

        self._value = self.instrument.get(self.knob)

        if isinstance(self._value, numbers.Number):
            self._value = self.multiplier * self._value + self.offset

        self.last_evaluation = time.time()

        return self._value

    @value.setter
    @Variable.setter_type_validator
    def value(self, value):
        """
        Set an instrument knob to value
        """

        try:

            upper_limit_exceeded = (self.upper_limit is not None) and (
                value > self.upper_limit
            )

            lower_limit_exceeded = (self.lower_limit is not None) and (
                value < self.lower_limit
            )

            if upper_limit_exceeded:
                self.instrument.set(
                    self.knob, (self.upper_limit - self.offset) / self.multiplier
                )

            if lower_limit_exceeded:
                self.instrument.set(
                    self.knob, (self.lower_limit - self.offset) / self.multiplier
                )

            if not upper_limit_exceeded and not lower_limit_exceeded:
                if isinstance(value, numbers.Number):
                    self.instrument.set(
                        self.knob, (value - self.offset) / self.multiplier
                    )
                else:
                    self.instrument.set(self.knob, value)

        except TypeError as type_error:

            logger.warning(str(type_error))

        self._value = self.instrument.__getattribute__(self.knob.replace(" ", "_"))

        if isinstance(self._value, numbers.Number):
            self._value = self.multiplier * self._value + self.offset

    def __str__(self):
        return f"Knob({self.value})"


class Meter(Variable):
    """
    Variable that is measured by an instrument, such as temperature.

    Some meters can be controlled directly or indirectly through
    an associated (but distinct) knob.

    The `instrument` argument specifies which instrument the meter is
    associated with.

    The `meter` argument is the label of the meter on the instrument.

    The `gate` optional argument is another variable which gates measurements.
    This is useful for situations where indefinitely continuous measurement of
    the meter, as in the usual experiment loop, is not desirable. Generally,
    it should be a variable of integer, boolean or toggle type. When the gate
    variable evaluates to 1/True/On, the meter can be measured. Otherwise,
    attempts to measure the meter will have no effect (`None` is returned).

    The optional `multiplier` and `offset` keyword arguments provide a means to affect
    a linear transformation of the raw meter value. Readings from the instrument will
    be multiplied by the `multiplier` and then increased by the `offset`.
    """

    _settable = False  #:

    def __init__(
        self,
        instrument: Instrument,
        meter: str,
        gate=None,
        multiplier: typing.Union[float, int] = 1,
        offset: typing.Union[float, int] = 0,
    ):
        self.instrument = instrument
        self.meter = meter
        self.multiplier = multiplier
        self.offset = offset

        if gate and isinstance(gate, Variable):
            self.gate = gate
        else:
            self.gate = Parameter(ON)

        self._type = typing.get_type_hints(
            getattr(instrument, "measure_" + meter.replace(" ", "_"))
        ).get("return", np.float64)

        self._value = None

    @property
    @Variable.getter_type_validator
    def value(self):
        """
        Measured value of the meter of an instrument
        """

        if not self.gate.value:
            return None

        self._value = self.instrument.measure(self.meter)

        if isinstance(self._value, numbers.Number):
            self._value = self.multiplier * self._value + self.offset

        self.last_evaluation = time.time()

        return self._value

    def __str__(self):
        return f"Meter({self.value})"


class Expression(Variable):
    """
    Variable that is calculated based on other variables of the experiment

    For example, the output power of a power supply could be recorded as an
    expression, where voltage is a knob and current is a meter:
    power = voltage * current.

    The `expression` argument is a string that can be evaluated by the Python
    interpreter, replacing symbols with the values of variables according to
    the `definitions` argument.

    The `definitions` argument is a dictionary mapping symbols in the
    `expression` argument to variables.
    """

    _settable = False  #:

    # shorthand terms for common functions
    # TODO consolidate these functions in a separate module
    _functions = {
        "sqrt(": "np.sqrt(",
        "exp(": "np.exp(",
        "sin(": "np.sin(",
        "cos(": "np.cos(",
        "tan(": "np.tan(",
        "sum(": "np.nansum(",
        "mean(": "np.nanmean(",
        "rms(": "np.nanstd(",
        "std(": "np.nanstd(",
        "var(": "np.nanvar(",
        "diff(": "np.diff(",
        "max(": "np.nanmax(",
        "min(": "np.nanmin(",
        "fft(": "self.fft(",
        "ifft(": "self.ifft(",
        "carrier(": "self.carrier(",
        "ampl(": "self.ampl(",
        "demod(": "self.demod(",
    }

    def __init__(self, expression: str, definitions: dict = None):
        self.expression = expression
        self.definitions = definitions if definitions is not None else {}

    @property
    @Variable.getter_type_validator
    def value(self):
        """
        Value of the expression
        """

        expression = self.expression

        logger.debug(f"Evaluating expression {expression}")

        # carets represent exponents
        expression = expression.replace("^", "**")

        variables = {
            symbol: variable._value for symbol, variable in self.definitions.items()
        }

        for shorthand, longhand in self._functions.items():
            if shorthand in expression:
                expression = expression.replace(shorthand, longhand)

        try:
            all_values = np.concatenate(
                [np.atleast_1d(val).flatten() for val in variables.values()]
            )

            no_nones = None not in all_values
            no_nans = np.nan not in all_values
            no_infs = (np.inf not in all_values) and (-np.inf not in all_values)

            valid_values = no_nones and no_nans and no_infs

            if valid_values:
                self._value = eval(expression, {**globals(), **variables}, locals())
            else:

                log_str = f"Dependencies for {expression} contain invalid values: "

                for name, value in variables.items():

                    if value is None:
                        log_str += f"\n{name} = None"
                    elif np.isnan(value):
                        log_str += f"\n{name} = NaN"
                    elif np.isinf(value):
                        log_str += f"\n{name} = +/-Inf"

                logger.debug(log_str)

                self._value = None

        except Exception as err:

            logger.warning(
                f"Unable to evaluate expression {self.expression} due to error: {err}"
            )

            self._value = None

        self.last_evaluation = time.time()

        logger.debug(f"Expression {self.expression} evaluated to {self._value}")

        return self._value

    def __str__(self):
        if len(str(self.value)) < 100:  # first call to value evaluates expression
            return f"Expression({self.expression} = {self._value})"
        elif isinstance(self._value, Array):
            return f"Expression({self.expression} = Array{np.shape(self._value)}"
        else:
            return (
                f"Expression({self.expression} = "
                f"{str(self._value)[:50]} ... {str(self._value)[-50:]}"
            )

    # Utility functions for Fourier analysis
    @staticmethod
    def fft(s):
        """Calculate the fast Fourier transform of a signal"""
        return np.fft.fft(s, norm="forward")

    @staticmethod
    def ifft(s):
        """Calculate the inverse fast Fourier transform of a signal"""
        return np.fft.ifft(s, norm="forward")

    @staticmethod
    def _find_carrier(s, dt, f0=0.0, bw=np.inf):
        """Characterize the carrier wave of a signal"""
        fft_s = np.fft.fft(s, norm="forward")
        f = np.fft.fftfreq(len(s), d=dt)

        in_band = (f > f0 - 0.5 * bw) & (f < f0 + 0.5 * bw)

        filt_fft_s = np.abs(in_band * fft_s)

        fc = np.abs(f[filt_fft_s == np.max(filt_fft_s)][0])
        Ac = np.abs(filt_fft_s[filt_fft_s == np.max(filt_fft_s)][0])

        return fc, Ac

    @staticmethod
    def carrier(s, dt, f0=0.0, bw=np.inf):
        """Find the carrier frequency of a signal"""
        return Expression._find_carrier(s, dt, f0=f0, bw=bw)[0]

    @staticmethod
    def ampl(s, dt, f0=0.0, bw=np.inf):
        """Calculate the amplitude of the carrier wave in a signal"""
        return 2 * Expression._find_carrier(s, dt, f0=f0, bw=bw)[1]

    @staticmethod
    def demod(s, dt, f0, bw=np.inf, cycles=np.inf, filt=None):
        """Demodulate an oscillatory signal"""

        if np.isfinite(cycles):
            # Partition signal into segments containing integer number of carrier cycles
            fc = Expression.carrier(s, dt, f0, bw=bw)  # get carrier frequency

            k_partition = int(cycles / (dt * fc))  # number of samples per partition

            n_parts = int(len(s) / k_partition) + 1

            s_padded = np.concatenate([s, np.zeros(n_parts * k_partition - len(s))])

            partitions = np.reshape(s_padded, (n_parts, k_partition))
        else:
            # Demodulate the whole signal

            n_parts = 1
            k_partition = len(s)

            partitions = np.array([s])

        partitions_demod = np.empty_like(partitions, dtype=np.complex128)

        for i, partition in enumerate(partitions):
            # Calculate FFT
            freq = np.fft.fftfreq(k_partition, d=dt)  # frequency values for the FFT
            fft = np.fft.fft(partition)

            # Find principal frequency within the given band about the given frequency
            fft_in_positive_band = np.abs(
                fft * ((freq > f0 - 0.5 * bw) & (freq < f0 + 0.5 * bw))
            )

            where_f0_closest = np.argwhere(
                fft_in_positive_band == np.max(fft_in_positive_band)
            ).flatten()[0]

            if np.abs(fft[where_f0_closest]) > 0.0:
                phase = np.log(fft[where_f0_closest]).imag  # Get phase of sinusoid
            else:
                phase = 0.0

            # Construct the demodulated FFT
            fft_demod = np.zeros_like(fft)

            # roll the positive component towards zero and remove phase
            fft_demod += np.roll(fft, -where_f0_closest) * np.exp(-1j * phase)

            # roll the negative component towards zero and remove phase
            fft_demod += np.roll(fft, where_f0_closest) * np.exp(1j * phase)

            # Apply low pass filter
            if filt == "gaussian":
                fft_demod *= np.exp(-(freq**2) / (2 * bw**2))
            if filt == "sinc":
                fft_demod *= np.sinc(freq / bw)
            else:
                fft_demod *= np.abs(freq) < 0.5 * bw

            partitions_demod[i] = np.fft.ifft(fft_demod)

        signal_demod = partitions_demod.flatten()

        return np.abs(signal_demod)


class Remote(Variable):
    """
    Variable controlled by an experiment (running a server routine) on a
    different process or computer.

    The `server` argument is the IP address and port of the server,
    in the form '(ip address)::(port)'.

    The `alias` argument identifies the particular variable on the server to
    link to. For socket servers, this is simply the name of the variable on
    the server. For Modbus servers, this is the starting address of the
    holding or input register for a read/write or readonly variable,
    respectively. For either set of registers, the starting address is 5*(n-1)
    for the nth variable in the `knobs` or `meters` of variables in the server routine
    definition.

    The (optional) `protocol` argument indicates which kind of server to connect
    to. Setting `protocol='modbus'` indicates a Modbus server (controlled by a
    ModbusServer routine on the remote process/computer), and any other value
    or no value indicates a socket server (controlled by a SocketServer
    routine).

    The `settable` argument is required for remote variables on a Modbus server,
    and is not used for a socket server. If `settable` is set to True, then
    the variable value is read from the holding registers (`knobs`), otherwise the
    variable value is read from the input registers (`meters`).

    The optional `multiplier` and `offset` keyword arguments provide a means to affect
    a linear transformation of the raw variable value. Readings from the instrument will
    be multiplied by the `multiplier` and then increased by the `offset`. Set commands
    to the server will take the variable value, subtract the `offset` and divide by the
    `multiplier`.
    """

    type_map = {
        Toggle: "64bit_uint",
        Boolean: "64bit_uint",
        Integer: "64bit_int",
        Float: "64bit_float",
    }

    def __init__(
        self,
        server: str,
        alias: typing.Union[int, str],
        protocol: str = None,
        settable: bool = False,  # needed for modbus protocol
        lower_limit: typing.Union[float, int] = None,
        upper_limit: typing.Union[float, int] = None,
        multiplier: typing.Union[float, int] = 1,
        offset: typing.Union[float, int] = 0,
    ):
        self.server = server
        self.alias = alias
        self.protocol = protocol
        self.lower_limit = lower_limit
        self.upper_limit = upper_limit
        self.multiplier = multiplier
        self.offset = offset

        if protocol == "modbus":
            self._client = ModbusClient(server)
            self._settable = settable

        else:
            server_ip, server_port = server.split("::")

            self._socket = socket.socket(socket.AF_INET, socket.SOCK_STREAM)

            self._socket.connect((server_ip, int(server_port)))

            self.get_settable()
            self.get_type()

    @property
    @Variable.getter_type_validator
    def value(self):
        """
        Value of the remote variable on a server
        """

        if self._type is None:
            self.get_type()

        if self.protocol == "modbus":
            fcode = 3 if self.settable else 4

            if self._type is not None:

                logger.debug(
                    f"Retrieving value of type {self._type} "
                    f"starting at register {self.alias}"
                    f"from Modbus server at {self.server}..."
                )

                self._value = self._client.read(
                    fcode, self.alias, count=4, _type=self.type_map[self._type]
                )

                logger.debug(
                    f"Value retrieved starting at register {self.alias} "
                    f"from Modbus server at {self.server} is {self._value}"
                )

        else:
            write_to_socket(self._socket, f"{self.alias} ?")

            logger.debug(
                f"Retrieving value of type {self._type} "
                f"with alias {self.alias}"
                f"from socket server at {self.server}..."
            )

            response = read_from_socket(self._socket, timeout=60, decode=False)

            try:
                if response is None:
                    self._value = None
                elif b"Error" in response:
                    raise RuntimeError(response.decode().split("Error: ")[-1])
                else:
                    bytes_value = response.split(self.alias.encode() + b" ")[-1].strip()

                    if bytes_value[:5] == b"dlpkl":
                        # pickled quantity, usually an array, list or tuple
                        self._value = dill.loads(bytes_value[5:])
                    else:
                        self._value = recast(
                            bytes_value,
                            to=self._type if self._type is not None else Type,
                        )

                logger.debug(
                    f"Value with alias {self.alias} retrieved "
                    f"from socket server at {self.server} is {self._value}"
                )

            except Exception as error:
                logger.warning(
                    f"Unable to retrieve value of {self.alias} "
                    f'from {self.protocol} server at {self.server}: "{error}"'
                )

        if isinstance(self._value, numbers.Number):
            self._value = self.multiplier * self._value + self.offset

        return self._value

    @value.setter
    @Variable.setter_type_validator
    def value(self, value):
        """
        Set the value of a remote variable
        """

        if isinstance(value, np.integer):
            value = (value - int(self.offset)) // int(self.multiplier)
        elif isinstance(value, np.floating):
            value = (value - self.offset) / self.multiplier

        if self.protocol == "modbus":

            logger.info(
                f"Writing value {value} to variable starting at register {self.alias}"
                f"on Modbus server at {self.server}..."
            )

            self._client.write(16, self.alias, value, _type=self.type_map[self._type])

        else:

            logger.debug(
                f"Writing value {value} to variable with alias {self.alias} "
                f"on socket server at {self.server}..."
            )

            write_to_socket(self._socket, f"{self.alias} {value}")

            check = read_from_socket(self._socket, timeout=60)

            if check == "" or check is None:
                logger.warning(
                    f"Received no response from server at "
                    f"{self.server} while trying to set {self.alias}"
                )
            elif "Error" in check:
                logger.warning(
                    f'Got response "{check}" while trying to set '
                    f"{self.alias} on server at {self.server}"
                )
            else:
                try:
                    check_value = recast(check.split(f"{self.alias} ")[1])

                    if value != check_value:
                        logger.warning(
                            f"Attempted to set {self.alias} on "
                            f"server at {self.server} to {value} but "
                            f"checked value is {check_value}"
                        )

                except ValueError as val_err:
                    logger.warning(
                        f"Unable to check value while setting "
                        f"{self.alias} on server at {self.server}; "
                        f'got error "{val_err}"'
                    )
                except IndexError as ind_err:
                    logger.warning(
                        f"Unable to check value while setting "
                        f"{self.alias} on server at {self.server}; "
                        f'got error "{ind_err}"'
                    )

    def __del__(self):
        if self.protocol == "modbus":
            self._client.disconnect()
        else:
            self._socket.shutdown(socket.SHUT_RDWR)
            self._socket.close()

    def get_type(self):
        """Get the data type of the remote variable"""

        if self.protocol == "modbus":

            logger.debug(
                f"Getting data type of variable starting at register {self.alias}"
                f"on Modbus server at {self.server}..."
            )

            fcode = 3 if self.settable else 4

            type_int = self._client.read(fcode, self.alias + 4, _type="16bit_int")

            self._type = {
                0: Boolean,
                1: Toggle,
                2: Integer,
                3: Float,
            }.get(type_int, None)

            logger.debug(
                f"Data type of variable starting at register {self.alias} "
                f"on Modbus server at {self.server} is {self._type}"
            )

        else:

            logger.debug(
                f"Getting data type of variable with alias {self.alias}"
                f"on socket server at {self.server}..."
            )

            write_to_socket(self._socket, f"{self.alias} type?")

            response = read_from_socket(self._socket, timeout=60)

            if response is not None:
                for _type in supported_types:
                    if str(_type) in response.split(self.alias)[-1]:
                        self._type = supported_types.get(_type, None)
            else:
                self._type = None

            logger.debug(
                f"Data type of variable with alias {self.alias} "
                f"on socket server at {self.server} is {self._type}"
            )

    def get_settable(self):
        """Get settability of remote variable"""
        write_to_socket(self._socket, f"{self.alias} settable?")

        response = read_from_socket(self._socket, timeout=60)
        self._settable = response == f"{self.alias} settable"

    def __str__(self):
        return f"Remote({self.alias}@{self.server} = {self.value})"


class Parameter(Variable):
    """
    Variable whose value is assigned directly by the user or indirectly with a
    routine. An example is a unit conversion factor such as 2.54 cm per inch, a
    numerical constant like pi or a setpoint for a control routine.

    The `parameter` argument is the given value of the parameter.
    """

    _settable = True  #:

    def __init__(
        self, parameter: typing.Union[float, int, bool, str, Toggle, np.ndarray]
    ):
        self._value = recast(parameter)

        for name, _type in supported_types.items():
            if isinstance(self._value, _type) and _type is not Type:

                self._type = _type

        logger.debug(f"Setting data type of parameter {self._value} to {self._type}")

    @property
    @Variable.getter_type_validator
    def value(self):
        """Value of the parameter"""
        return self._value

    @value.setter
    @Variable.setter_type_validator
    def value(self, value):
        """Set the parameter value"""
        self._value = value

    def __str__(self):
        return f"Parameter({self.value})"


supported = {
    key: value
    for key, value in vars().items()
    if type(value) is type and issubclass(value, Variable)
}



import re
import numpy as np

from empyric.tools import logger
from empyric.types import Float, Integer, String, Toggle, ON, OFF
from empyric.adapters import Modbus, ModbusSerial, Socket
from empyric.collection.instrument import Instrument, setter, getter, measurer


class OmegaCN7500(Instrument):
    """
    Omega model CN7500 PID temperature controller
    """

    name = "OmegaCN7500"

    supported_adapters = (
        (
            Modbus,
            {"baud_rate": 38400, "parity": "N", "delay": 0.2, "slave_id": 1},
        ),
    )

    scale_factors = {
        0: 0.1,  # K-Type Thermocouple
        1: 0.1,  # J-Type Thermocouple
        2: 0.1,  # T-Type Thermocouple
        3: 0.1,  # E-Type Thermocouple
        4: 0.1,  # N-Type Thermocouple
        5: 0.1,  # R-Type Thermocouple
        6: 0.1,  # S-Type Thermocouple
        7: 0.1,  # B-Type Thermocouple
        8: 0.1,  # L-Type Thermocouple
        9: 0.1,  # U-Type Thermocouple
        10: 0.1,  # TXK-Type Thermocouple
        11: 0.1,  # JPt100-Type RTD
        12: 0.1,  # Pt100-Type RTD
        13: 0.001,  # 0-5V Analog Input
        14: 0.001,  # 0-10V Analog Input
        15: 0.001,  # 0-20mA Analog Input
        16: 0.001,  # 4-20mA Analog Input
        17: 0.001,  # 0-50mV Analog Input
    }

    knobs = (
        "output",
        "setpoint",
        "proportional band",
        "integration time",
        "derivative time",
        "sensor type",
    )

    meters = ("temperature", "power", "output level")

    @setter
    def set_output(self, state: Toggle):
        if state == ON:
            # turn on output & start PID control
            self.write(5, 0x0814, 1)
        elif state == OFF:
            # turn off output & stop PID control
            self.write(5, 0x0814, 0)

    @getter
    def get_output(self) -> Toggle:
        result = self.read(1, 0x0814)
        if result:
            return ON
        else:
            return OFF

    @setter
    def set_setpoint(self, setpoint: Float):
        if self.get_sensor_type() in self.scale_factors.keys():
            scaler = self.scale_factors[self.get_sensor_type()]
        else:
            scaler = 0.1

        self.write(6, 0x1001, int(setpoint / scaler))

    @getter
    def get_setpoint(self) -> Float:
        if self.get_sensor_type() in self.scale_factors.keys():
            scaler = self.scale_factors[self.get_sensor_type()]
        else:
            scaler = 0.1

        return self.read(3, 0x1001) * scaler

    @setter
    def set_proportional_band(self, P: Integer):
        round_P = round(P)
        if P != round_P:
            logger.warning(f"Proportional band value {P} will be rounded to {round_P}")
        self.write(6, 0x1009, round_P)

    @getter
    def get_proportional_band(self) -> Integer:
        return self.read(3, 0x1009)

    @setter
    def set_integration_time(self, Ti: Integer):
        round_Ti = round(Ti)
        if Ti != round_Ti:
            logger.warning(f"Integration time value {Ti} will be rounded to {round_Ti}")
        self.write(6, 0x100A, round_Ti)

    @getter
    def get_integration_time(self) -> Integer:
        return self.read(3, 0x100A)

    @setter
    def set_derivative_time(self, Td: Integer):
        round_Td = round(Td)
        if Td != round_Td:
            logger.warning(f"Derivative time value {Td} will be rounded to {round_Td}")
        self.write(6, 0x100B, round_Td)

    @getter
    def get_derivative_time(self) -> Integer:
        return self.read(3, 0x100B)

    @setter
    def set_sensor_type(self, sensor_type: Integer):
        self.write(6, 0x1004, sensor_type)

    @getter
    def get_sensor_type(self) -> Integer:
        return self.read(3, 0x1004)

    @measurer
    def measure_temperature(self) -> Float:
        if self.get_sensor_type() in self.scale_factors.keys():
            scaler = self.scale_factors[self.get_sensor_type()]
        else:
            scaler = 0.1
        return self.read(3, 0x1000) * scaler

    @measurer
    def measure_power(self) -> Float:
        if self.get_sensor_type() in self.scale_factors.keys():
            scaler = self.scale_factors[self.get_sensor_type()]
        else:
            scaler = 0.1
        return self.read(3, 0x1000) * scaler

    @getter
    def measure_output_level(self) -> Float:
        if self.get_sensor_type() in self.scale_factors.keys():
            scaler = self.scale_factors[self.get_sensor_type()]
        else:
            scaler = 0.1
        return self.read(3, 0x1012) * scaler


class OmegaPlatinum(Instrument):
    """
    Omega Platinum series PID controller
    """

    name = "OmegaPlatinum"

    supported_adapters = ((ModbusSerial, {"baud_rate": 19200, "parity": "O"}),)

    knobs = (
        "setpoint",
        "autotune",
        "output",
        "tc type",
    )

    meters = ("temperature", "power")

    TC_map = {
        "K": 1,
        "J": 0,
        "T": 2,
        "E": 3,
        "N": 4,
        "R": 6,
        "S": 7,
        "B": 8,
        "C": 9,
    }

    inverse_TC_map = {v: k for k, v in TC_map.items()}

    @measurer
    def measure_temperature(self) -> Float:
        return self.read(0x0280, dtype="float")

    @measurer
    def measure_power(self) -> Float:
        return self.read(0x022A, dtype="float")

    @setter
    def set_setpoint(self, setpoint: Float):
        self.write(0x02E2, setpoint, dtype="float")

    @getter
    def get_setpoint(self) -> Float:
        return self.read(0x02E2, dtype="float")

    @setter
    def set_autotune(self, state: Toggle):
        if state == ON:
            self.write(0x0243, 1)
        else:
            self.write(0x0243, 0)

    @setter
    def set_output(self, state: Toggle):
        if state == ON:
            self.write(0x0240, 6)
        else:
            self.write(0x0240, 8)

    @getter
    def get_output(self) -> Toggle:
        state = self.read(0x0240)
        if state == 6 or state == 4:
            return ON
        else:
            return OFF

    @setter
    def set_tc_type(self, _type: String):
        try:
            self.write(0x0283, OmegaPlatinum.TC_map[_type.upper()])
        except (KeyError, AttributeError):
            raise ValueError(f"{self.name}: Invalid thermocouple type {_type}")

    @getter
    def get_tc_type(self) -> String:
        try:
            return self.inverse_TC_map[self.read(0x0283)]
        except KeyError:
            return "None"


class RedLionPXU(Instrument):
    """
    Red Lion PXU temperature PID controller
    """

    name = "RedLionPXU"

    supported_adapters = ((ModbusSerial, {"buad_rate": 38400}),)

    knobs = ("output", "setpoint", "autotune")

    meters = ("temperature", "power")

    @setter
    def set_output(self, state: Toggle):
        if state == ON:
            # turn on output & start PID control
            self.backend.write_bit(0x11, 1)
        elif state == OFF:
            # turn off output & stop PID control
            self.backend.write_bit(0x11, 0)

    @setter
    def set_setpoint(self, setpoint: Integer):
        self.write(0x1, int(setpoint))

    @measurer
    def measure_temperature(self) -> Integer:
        return self.read(0x0)

    @measurer
    def measure_power(self) -> Float:
        return self.read(0x8) / 10

    @setter
    def set_autotune(self, state: Toggle):
        if state == ON:
            self.write(0xF, 1)
        elif state == OFF:
            self.write(0xF, 0)


class WatlowEZZone(Instrument):
    """
    Watlow EZ-Zone PID process controller
    """

    name = "WatlowEZZone"

    supported_adapters = ((ModbusSerial, {"baud_rate": 9600}),)

    knobs = ("setpoint",)

    meters = ("temperature",)

    @measurer
    def measure_temperature(self) -> Float:
        # swapped little-endian byte order (= 3 in minimalmodbus)
        return self.read(360, dtype="float", byte_order=3)

    @getter
    def get_setpoint(self) -> Float:
        return self.read(2160, dtype="float", byte_order=3)

    @setter
    def set_setpoint(self, setpoint: Float):
        return self.write(2160, setpoint, dtype="float", byte_order=3)

    @getter
    def get_proportional_band(self) -> Float:
        return self.read(1890, dtype="float", byte_order=3)

    @setter
    def set_proportional_band(self, band: Float):
        return self.write(1890, band, dtype="float", byte_order=3)

    @getter
    def get_time_integral(self) -> Float:
        return self.read(1894, dtype="float", byte_order=3)

    @setter
    def set_time_integral(self, integral: Float):
        return self.write(1894, integral, dtype="float", byte_order=3)

    @getter
    def get_time_derivative(self) -> Float:
        return self.read(1896, dtype="float", byte_order=3)

    @setter
    def set_time_derivative(self, derivative: Float):
        return self.write(1896, derivative, dtype="float", byte_order=3)


class MKSGSeries(Instrument):
    """MKS G series mass flow controller"""

    name = "MKSGSeries"

    supported_adapters = ((Modbus, {}),)

    knobs = (
        "setpoint",  # flow rate setpoint in SCCM
        "ramp time",  # ramp time in milliseconds
    )

    meters = (
        "flow rate",  # actual flow rate in SCCM
        "valve position",  # valve position in percent
        "temperature",  # temperature in degrees C
    )

    @setter
    def set_setpoint(self, setpoint: Float):
        self.write(16, 0xA000, setpoint, _type="32bit_float")

    @getter
    def get_setpoint(self) -> Float:
        return self.read(3, 0xA000, count=2, _type="32bit_float")

    @setter
    def set_ramp_time(self, ramp_time: Integer):
        self.write(16, 0xA002, int(ramp_time), _type="32bit_uint")

    @getter
    def get_ramp_time(self) -> Integer:
        return self.read(3, 0xA002, count=2, _type="32bit_uint")

    @measurer
    def measure_flow_rate(self) -> Float:
        return self.read(4, 0x4000, count=2, _type="32bit_float")

    @measurer
    def measure_valve_position(self) -> Float:
        return self.read(4, 0x4004, count=2, _type="32bit_float")

    @measurer
    def measure_temperature(self) -> Float:
        return self.read(4, 0x4002, count=2, _type="32bit_float")


class AlicatMFC(Instrument):
    """Alicat mass flow controller"""

    name = "AlicatMFC"

    supported_adapters = ((Modbus, {}),)

    knobs = ("setpoint",)  # flow rate setpoint in SCCM

    meters = (
        "flow rate",  # actual flow rate in SCCM
        "temperature",  # temperature in degrees C
        "pressure",
        # pressure in PSI (absolute, gauge or differential,
        # depending on device configuration)
    )

    @setter
    def set_setpoint(self, setpoint: Float):
        self.write(16, 1009, setpoint, _type="32bit_float")

    @getter
    def get_setpoint(self) -> Float:
        return self.read(3, 1009, count=2, _type="32bit_float")

    @measurer
    def measure_flow_rate(self) -> Float:
        return self.read(4, 1208, count=2, _type="32bit_float")

    @measurer
    def measure_temperature(self) -> Float:
        return self.read(4, 1204, count=2, _type="32bit_float")

    @measurer
    def measure_pressure(self) -> Float:
        return self.read(4, 1202, count=2, _type="32bit_float")


class SynaccessNetbooter(Instrument):
    supported_adapters = (
        (Socket, {"write_termination": "\r\n", "read_termination": None}),
    )

    knobs = (
        "port 1 toggle",
        "port 2 toggle",
        "port 3 toggle",
        "port 4 toggle",
        "port 5 toggle",
        "port 6 toggle",
        "port 7 toggle",
        "port 8 toggle",
        "port 9 toggle",
        "port 10 toggle",
        "port 11 toggle",
        "port 12 toggle",
        "port 13 toggle",
        "port 14 toggle",
        "port 15 toggle",
        "port 16 toggle",
    )

    def _set_port_n_toggle(self, n, state):
        if state != ON and state != OFF:
            raise ValueError(
                f"port toggle state of {self.name} must be "
                f"either ON or OFF (Toggle type)"
            )

        if state == ON:
            self.write("$A3 %d 1" % n)
        elif state == OFF:
            self.write("$A3 %d 0" % n)

    def _get_port_n_toggle(self, n):
        # Dump buffer (this device sends out a Telnet handshake upon initial
        # connection and periodically transmits null bytes, possibly as a
        # keep-alive signal)
        self.read(nbytes=np.inf, timeout=0.1, decode=False)

        def termination(message):
            return re.search(b"A0,\d+", message)

        status_message = self.query("$A5", termination=termination, decode=False)

        # Port statuses are a sequence of 0s and 1s, starting from the right
        statuses = re.search(b"A0,\d+", status_message)[0]

        port_n_toggle = ON if int(statuses.decode()[-n]) == 1 else OFF

        return port_n_toggle

    @setter
    def set_port_1_toggle(self, state: Toggle):
        return self._set_port_n_toggle(1, state)

    @getter
    def get_port_1_toggle(self) -> Toggle:
        return self._get_port_n_toggle(1)

    @setter
    def set_port_2_toggle(self, state: Toggle):
        return self._set_port_n_toggle(2, state)

    @getter
    def get_port_2_toggle(self) -> Toggle:
        return self._get_port_n_toggle(2)

    @setter
    def set_port_3_toggle(self, state: Toggle):
        return self._set_port_n_toggle(3, state)

    @getter
    def get_port_3_toggle(self) -> Toggle:
        return self._get_port_n_toggle(3)

    @setter
    def set_port_4_toggle(self, state: Toggle):
        return self._set_port_n_toggle(4, state)

    @getter
    def get_port_4_toggle(self) -> Toggle:
        return self._get_port_n_toggle(4)

    @setter
    def set_port_5_toggle(self, state: Toggle):
        return self._set_port_n_toggle(5, state)

    @getter
    def get_port_5_toggle(self) -> Toggle:
        return self._get_port_n_toggle(5)

    @setter
    def set_port_6_toggle(self, state: Toggle):
        return self._set_port_n_toggle(6, state)

    @getter
    def get_port_6_toggle(self) -> Toggle:
        return self._get_port_n_toggle(6)

    @setter
    def set_port_7_toggle(self, state: Toggle):
        return self._set_port_n_toggle(7, state)

    @getter
    def get_port_7_toggle(self) -> Toggle:
        return self._get_port_n_toggle(7)

    @setter
    def set_port_8_toggle(self, state: Toggle):
        return self._set_port_n_toggle(8, state)

    @getter
    def get_port_8_toggle(self) -> Toggle:
        return self._get_port_n_toggle(8)

    @setter
    def set_port_9_toggle(self, state: Toggle):
        return self._set_port_n_toggle(9, state)

    @getter
    def get_port_9_toggle(self) -> Toggle:
        return self._get_port_n_toggle(9)

    @setter
    def set_port_10_toggle(self, state: Toggle):
        return self._set_port_n_toggle(10, state)

    @getter
    def get_port_10_toggle(self) -> Toggle:
        return self._get_port_n_toggle(10)

    @setter
    def set_port_11_toggle(self, state: Toggle):
        return self._set_port_n_toggle(11, state)

    @getter
    def get_port_11_toggle(self) -> Toggle:
        return self._get_port_n_toggle(11)

    @setter
    def set_port_12_toggle(self, state: Toggle):
        return self._set_port_n_toggle(12, state)

    @getter
    def get_port_12_toggle(self) -> Toggle:
        return self._get_port_n_toggle(12)

    @setter
    def set_port_13_toggle(self, state: Toggle):
        return self._set_port_n_toggle(13, state)

    @getter
    def get_port_13_toggle(self) -> Toggle:
        return self._get_port_n_toggle(13)

    @setter
    def set_port_14_toggle(self, state: Toggle):
        return self._set_port_n_toggle(14, state)

    @getter
    def get_port_14_toggle(self) -> Toggle:
        return self._get_port_n_toggle(14)

    @setter
    def set_port_15_toggle(self, state: Toggle):
        return self._set_port_n_toggle(15, state)

    @getter
    def get_port_15_toggle(self) -> Toggle:
        return self._get_port_n_toggle(15)

    @setter
    def set_port_16_toggle(self, state: Toggle):
        return self._set_port_n_toggle(16, state)

    @getter
    def get_port_16_toggle(self) -> Toggle:
        return self._get_port_n_toggle(16)



import re

from empyric.types import recast, Float
from empyric.adapters import Socket, Modbus
from empyric.collection.instrument import Instrument, setter, getter, measurer


class BrainboxesED560(Instrument):
    """
    Brainboxes 4 channel analog output (0-10 V / 0-20mA) gateway.

    For socket communication, the default port is 9500. If the IP address is
    unknown, you can use the Boost.IO Driver software to find it.

    ASCII Protocol must be used.
    """

    supported_adapters = ((Socket, {}),)

    knobs = (
        "analog_out0",
        "analog_out1",
        "analog_out2",
        "analog_out3",
    )

    @setter
    def set_analog_out0(self, value: Float):
        self.query("#010%f" % float(value), validator=self._set_validator)

    @getter
    def get_analog_out0(self) -> Float:
        response = self.query("$0160", validator=self._get_validator)
        return recast(response[3:])

    @setter
    def set_analog_out1(self, value: Float):
        self.query("#011%f" % float(value), validator=self._set_validator)

    @getter
    def get_analog_out1(self) -> Float:
        response = self.query("$0161", validator=self._get_validator)
        return recast(response[3:])

    @setter
    def set_analog_out2(self, value: Float):
        self.query("#012%f" % float(value), validator=self._set_validator)

    @getter
    def get_analog_out2(self) -> Float:
        response = self.query("$0162", validator=self._get_validator)
        return recast(response[3:])

    @setter
    def set_analog_out3(self, value: Float):
        self.query("#013%f" % float(value), validator=self._set_validator)

    @getter
    def get_analog_out3(self) -> Float:
        response = self.query("$0163", validator=self._get_validator)
        return recast(response[3:])

    @staticmethod
    def _get_validator(response):
        return re.match("!01\+\d\d\.\d\d\d", response)

    @staticmethod
    def _set_validator(response):
        return response.strip() == ">"


class BrainboxesED549(Instrument):
    """
    Brainboxes 4 channel analog input (0-10 V / 0-20mA) gateway.

    For socket communication, the default port is 9500. If the IP address is
    unknown, you can use the Boost.IO Driver software to find it.

    ASCII Protocol must be used.
    """

    supported_adapters = ((Socket, {}),)

    meters = (
        "analog_in0",
        "analog_in1",
        "analog_in2",
        "analog_in3",
    )

    @measurer
    def measure_analog_in0(self) -> Float:
        response = self.query("#010", validator=self._validator)
        return response[1:]

    @measurer
    def measure_analog_in1(self) -> Float:
        response = self.query("#011", validator=self._validator)
        return response[1:]

    @measurer
    def measure_analog_in2(self) -> Float:
        response = self.query("#012", validator=self._validator)
        return response[1:]

    @measurer
    def measure_analog_in3(self) -> Float:
        response = self.query("#013", validator=self._validator)
        return response[1:]

    @staticmethod
    def _validator(response):
        return re.match(">\+?\-?\d\d\.\d\d\d", response)


class ADAM6024(Instrument):
    """
    Advantech ADAM-6024 6 channel analog input (0-10 V / 0-20mA) / 2 channel
    analog output module.

    Using Modbus Protocol
    """

    supported_adapters = ((Modbus, {"byte_order": ">"}),)

    knobs = (
        "analog_out0",
        "analog_out1",
        "analog_in0_type",
        "analog_in1_type",
        "analog_in2_type",
        "analog_in3_type",
        "analog_in4_type",
        "analog_in5_type",
        "analog_out0_type",
        "analog_out1_type",
    )

    meters = (
        "analog_in0",
        "analog_in1",
        "analog_in2",
        "analog_in3",
        "analog_in4",
        "analog_in5",
    )

    presets = {
        "analog_in0_type": "+/-10V",
        "analog_in1_type": "+/-10V",
        "analog_in2_type": "+/-10V",
        "analog_in3_type": "+/-10V",
        "analog_in4_type": "+/-10V",
        "analog_in5_type": "+/-10V",
        "analog_out0_type": "0~10V",
        "analog_out1_type": "0~10V",
    }

    AIN_TYPES = {"4~20mA": 0x07, "+/-10V": 0x08, "0~20mA": 0x0D}

    AOUT_TYPES = {"0~20mA": 0x00, "4~20mA": 0x01, "0~10V": 0x02}

    def _measure_AIN(self, n) -> Float:
        raw = self.read(4, n, _type="16bit_uint")
        match self._get_AIN_type(n):
            case "4~20mA":
                return raw / 65535.0 * 16 + 4
            case "+/-10V":
                return raw / 65535.0 * 20 - 10
            case "0~20mA":
                return raw / 65535.0 * 20
            case _:
                raise TypeError("Unrecognized Type")
                return None

    def _set_AOUT(self, n, value) -> Float:
        match self._get_AOUT_type(n):
            case "0~20mA":
                scaled_value = int((value) / 20.0 * ((2**12) - 1))
            case "4~20mA":
                scaled_value = int((value - 4) / 16.0 * ((2**12) - 1))
            case "0~10V":
                scaled_value = int((value) / 10.0 * ((2**12) - 1))
            case _:
                scaled_value = None
                raise TypeError("Unrecognized Type")
        return self.write(6, 10 + n, scaled_value, _type="16bit_uint")

    def _get_AOUT(self, n) -> Float:
        raw = self.read(4, 10 + n, _type="16bit_uint")
        match self._get_AOUT_type(n):
            case "0~20mA":
                return raw / float(2**12 - 1) * 20
            case "4~20mA":
                return raw / float(2**12 - 1) * 16 + 4
            case "0~10V":
                return raw / float(2**12 - 1) * 10
            case _:
                raise TypeError("Unrecognized Type")
                return None

    def _set_AIN_type(self, n, str_value: str) -> Float:
        if str_value in self.AIN_TYPES.keys():
            value = self.AIN_TYPES[str_value]
            return self.write(6, 200 + n, value, _type="16bit_uint")
        else:
            raise TypeError(f"Value not in [{self.AIN_TYPES}]")

    def _get_AIN_type(self, n) -> str:
        ain_types_rev = {val: key for key, val in self.AIN_TYPES.items()}
        return ain_types_rev[self.read(4, 200 + n, _type="16bit_uint")]

    def _set_AOUT_type(self, n, str_value) -> Float:
        if str_value in self.AOUT_TYPES.keys():
            value = self.AOUT_TYPES[str_value]
            return self.write(6, 208 + n, value, _type="16bit_uint")
        else:
            raise TypeError(f"Value not in [{self.AOUT_TYPES}]")

    def _get_AOUT_type(self, n) -> str:
        aout_types_rev = {val: key for key, val in self.AOUT_TYPES.items()}
        return aout_types_rev[self.read(4, 208 + n, _type="16bit_uint")]

    @setter
    def set_analog_out0(self, value: Float):
        self._set_AOUT(0, value)

    @getter
    def get_analog_out0(self) -> Float:
        return self._get_AOUT(0)

    @setter
    def set_analog_out1(self, value: Float):
        self._set_AOUT(1, value)

    @getter
    def get_analog_out1(self) -> Float:
        return self._get_AOUT(1)

    @measurer
    def measure_analog_in0(self) -> Float:
        return self._measure_AIN(0)

    @measurer
    def measure_analog_in1(self) -> Float:
        return self._measure_AIN(1)

    @measurer
    def measure_analog_in2(self) -> Float:
        return self._measure_AIN(2)

    @measurer
    def measure_analog_in3(self) -> Float:
        return self._measure_AIN(3)

    @measurer
    def measure_analog_in4(self) -> Float:
        return self._measure_AIN(4)

    @measurer
    def measure_analog_in5(self) -> Float:
        return self._measure_AIN(5)

    @setter
    def set_analog_out0_type(self, value: str):
        self._set_AOUT_type(0, value)

    @getter
    def get_analog_out0_type(self) -> str:
        return self._get_AOUT_type(0)

    @setter
    def set_analog_out1_type(self, value: str):
        self._set_AOUT_type(1, value)

    @getter
    def get_analog_out1_type(self) -> str:
        return self._get_AOUT_type(1)

    @setter
    def set_analog_in0_type(self, value: str):
        self._set_AIN_type(0, value)

    @getter
    def get_analog_in0_type(self) -> str:
        return self._get_AIN_type(0)

    @setter
    def set_analog_in1_type(self, value: str):
        self._set_AIN_type(1, value)

    @getter
    def get_analog_in1_type(self) -> str:
        return self._get_AIN_type(1)

    @setter
    def set_analog_in2_type(self, value: str):
        self._set_AIN_type(2, value)

    @getter
    def get_analog_in2_type(self) -> str:
        return self._get_AIN_type(2)

    @setter
    def set_analog_in3_type(self, value: str):
        self._set_AIN_type(3, value)

    @getter
    def get_analog_in3_type(self) -> str:
        return self._get_AIN_type(3)

    @setter
    def set_analog_in4_type(self, value: str):
        self._set_AIN_type(4, value)

    @getter
    def get_analog_in4_type(self) -> str:
        return self._get_AIN_type(4)

    @setter
    def set_analog_in5_type(self, value: str):
        self._set_AIN_type(5, value)

    @getter
    def get_analog_in5_type(self) -> str:
        return self._get_AIN_type(5)



import typing
from threading import RLock
from functools import wraps

from empyric.tools import logger
from empyric.types import recast, Type, ON, OFF, Toggle
from empyric.adapters import Adapter, AdapterError


def setter(method):
    """
    Utility function that wraps all set_[knob] methods and records the new
    knob values.

    If the wrapped method returns a value, this value is assigned to the
    corresponding knob attribute of the instrument. This is convenient for
    cases where the set value maps to another value which is more relevant.
    Otherwise, the value of the method's first argument is assigned to the
    attribute.

    :param method: (callable) method to be wrapped

    :return: wrapped method
    """

    knob = "_".join(method.__name__.split("_")[1:])

    type_hints = typing.get_type_hints(method)
    type_hints.pop("self", None)
    type_hints.pop("return", None)

    if type_hints:
        dtype = type_hints[list(type_hints)[0]]
    else:
        dtype = Type

    logger.debug(f'Knob {" ".join(knob.split("_"))} dtype is {dtype}')

    @wraps(method)
    def wrapped_method(*args, **kwargs):

        self = args[0]
        value = args[1]

        if not self.adapter.connected and knob != "connected":
            logger.warning(
                f"Instrument {self.name} is disconnected; unable to set {knob}"
            )
            self.__setattr__(knob, None)
            return

        self.lock.acquire()

        try:

            args = list(args)
            args[1] = recast(args[1], to=dtype)

            logger.debug(f"Setting {knob} on {self.name} to {value}")

            returned_value = method(*args, **kwargs)

            # The knob attribute is set to the returned value of the method, or
            # the value argument if the returned value is None
            if returned_value is not None:

                logger.debug(
                    f"Value returned by set method for {knob} on {self.name} "
                    f"is {returned_value} instead of applied value {value}"
                )

                self.__setattr__(knob, recast(returned_value, to=dtype))
            else:
                self.__setattr__(knob, recast(value, to=dtype))
        finally:
            self.lock.release()

    return wrapped_method


def getter(method):
    """
    Utility function that wraps all get_[knob] methods and records the
    retrieved knob values.

    :param method: (callable) method to be wrapped

    :return: wrapped method
    """

    knob = "_".join(method.__name__.split("_")[1:])

    dtype = typing.get_type_hints(method).get("return", Type)

    logger.debug(f'Knob {" ".join(knob.split("_"))} dtype is {dtype}')

    @wraps(method)
    def wrapped_method(*args, **kwargs):
        self = args[0]

        if not self.adapter.connected and knob != "connected":
            self.__setattr__(knob, None)
            logger.warning(
                f"Instrument {self.name} is disconnected; unable to get {knob}"
            )
            return

        self.lock.acquire()

        try:

            logger.debug(f"Getting value of {knob} on {self.name}...")

            value = recast(method(*args, **kwargs), to=dtype)

            logger.debug(f"Retrieved value of {knob} on {self.name} is {value}")
        except AttributeError as err:
            # catches most errors caused by the adapter returning None
            if "NoneType" in str(err):

                logger.debug(
                    f"Unable to retrieve non-null value for {knob} on {self.name}"
                )

                value = None
            else:
                raise AttributeError(err)
        finally:
            self.lock.release()

        self.__setattr__(knob, value)

        return value

    return wrapped_method


def measurer(method):
    """
    Utility function that wraps all measure_[meter] methods and records the
    measured value.

    :param method: (callable) method to be wrapped

    :return: wrapped method
    """

    meter = "_".join(method.__name__.split("_")[1:])

    dtype = typing.get_type_hints(method).get("return", Type)

    logger.debug(f'Meter {" ".join(meter.split("_"))} dtype is {dtype}')

    @wraps(method)
    def wrapped_method(*args, **kwargs):
        self = args[0]

        if not self.adapter.connected:
            self.__setattr__(meter, None)
            print(f"Instrument {self.name} is disconnected; unable to measure {meter}")
            return

        self.lock.acquire()

        try:

            logger.debug(f"Measuring value of {meter} on {self.name}...")

            value = recast(method(*args, **kwargs), to=dtype)

            logger.debug(f"Measured value of {meter} on {self.name} is {value}")
        except AttributeError as err:
            # catches most errors caused by the adapter returning None
            if "NoneType" in str(err):

                logger.debug(
                    f"Unable to measure non-null value for {meter} on {self.name}"
                )

                value = None
            else:
                raise AttributeError(err)
        finally:
            self.lock.release()

        self.__setattr__(meter, value)

        return value

    return wrapped_method


class Instrument:
    """
    Basic representation of an instrument, essentially a set of knobs and meters

    Each instrument has the following attributes:

    * ``name``: the name of the instrument. Once connected via  an adapter,
      this is converted to
      ``name + '@' + address``.
    * ``supported adapters``: tuple of 2-element tuples. Each 2-element tuple
      contains an adapter class that the instrument can be used with and a
      dictionary of adapter settings.
    * ``knobs``: tuple of the names of all knobs that can be set on the
      instrument. Every instrument has a ``connected`` (``Toggle``) knob, for
      convenience, whose set method calls the instruments ``connect`` or ``disconnect``
      methods, if set to ``ON`` or ``OFF`` respectively.
    * ``presets``: dictionary of knob settings to apply when the instrument is
      instantiated.
      The keys are the names of the knobs and the values are the knob values.
    * ``postsets``: dictionary of knob settings (same format as ``presets``)
      to apply when the instrument is deleted.
    * ``meters``: tuple of the names of all meters that can be measured on
      this instrument.

    Every knob of an instrument has an associated ``set_[knob]`` method, which sends
    commands to the physical instrument that change the value of the knob to the given
    value. Each ``set_[knob]`` method should be wrapped by the ``setter`` function
    defined above.

    Every knob may also have an associated ``get_[knob]`` method, which queries the
    value of the knob from the physical instrument, and should be wrapped by the
    ``getter`` function defined above.

    Every meter of an instrument has an associated ``measure_[meter]`` method, which
    queries the value of a quantity measured by the physical instrument, and should be
    wrapped by the ``measurer`` function defined above.

    For convenience, the ``setter``, ``getter`` and ```measurer`` functions augment the
    corresponding methods primarily by enforcing typing on the method arguments and the
    returned values, and storing the last known values of knobs and meters in the
    corresponding attribute of the Instrument instance. For example, if a meter called
    ""temperature" is retrieved via the corresponding ``measure_temperature`` method of
    the instrument ``thermometer``, the wrapping ``measurer`` function will assign that
    temperature value to ``thermometer.temperature``. If the expected data type is a
    floating point number, the ``measurer`` function additionally converts whatever the
    bare ``measure_temperature`` method returns to a 64-bit floating point value.

    """

    name = "Instrument"

    supported_adapters = ((Adapter, {}),)

    knobs = tuple()

    presets = {}

    postsets = {}

    meters = tuple()

    ignore_errors = False

    def __init__(
        self, address=None, adapter=None, presets=None, postsets=None, **kwargs
    ):
        """

        :param address: (str/int) address of instrument
        :param adapter: (Adapter) desired adapter to use for communications
        with the instrument
        :param presets: (dict) dictionary of instrument presets of the form
        {..., knob: value, ...}
        :param presets: (dict) dictionary of instrument postsets of the form
        {..., knob: value, ...}
        :param kwargs: (dict) any keyword args for the adapter
        """

        self.address = address

        if "connected" not in self.knobs:
            self.knobs = ("connected",) + self.knobs

        adapter_connected = False
        if adapter:
            self.adapter = adapter(self, **kwargs)
        else:
            errors = []
            for _adapter, settings in self.supported_adapters:
                settings.update(kwargs)
                try:
                    self.adapter = _adapter(self, **settings)
                    adapter_connected = True
                    break
                except BaseException as error:
                    msg = (
                        f"in trying {_adapter.__name__} adapter, "
                        f"got {type(error).__name__}: {error}"
                    )
                    errors.append(msg)

            if not adapter_connected:
                message = (
                    f"unable to connect an adapter to "
                    f"instrument {self.name} at address {address}:\n"
                )
                for error in errors:
                    message = message + f"{error}\n"
                raise ConnectionError(message)

        if "@" not in self.name:
            if self.address:
                self.name = self.name + "@" + str(self.address)
            else:
                self.name = self.name + "@" + hex(id(self))

        # This lock is used to prevent commands executed in separate threads from
        # interfering with each other. The lock is acquired in the setter, getter
        # and measurer wrapper functions and then released when the wrapped
        # operation is complete. Using an RLock allows set, get and measure
        # methods to call other such methods without blocking, as long as it
        # happens in the same thread, which is the norm.
        self.lock = RLock()

        # Get existing knob settings, if possible
        for knob in self.knobs:
            if hasattr(self, "get_" + knob.replace(" ", "_")):
                # retrieves the knob value from the instrument
                self.__getattribute__("get_" + knob.replace(" ", "_"))()
            else:
                # knob value is unknown until it is set
                self.__setattr__(knob.replace(" ", "_"), None)

        # Apply presets
        if presets:
            self.presets = {**self.presets, **presets}

        for knob, value in self.presets.items():
            self.set(knob, value)

        # Get postsets
        if postsets:
            self.postsets = {**self.postsets, **postsets}

        self.kwargs = kwargs

    def __repr__(self):
        return self.name

    # map write, read and query methods to the adapter's
    def write(self, *args, **kwargs):
        """
        Alias for the adapter's write method

        :param args: any arguments for the adapter's write method, usually
                     including a command string
        :param kwargs: any arguments for the adapter's write method
        :return: whatever is returned by the adapter's write method
        """

        try:
            return self.adapter.write(*args, **kwargs)
        except AdapterError as adapter_error:
            if self.ignore_errors:
                logger.error(str(adapter_error))
            else:
                raise adapter_error

    def read(self, *args, **kwargs):
        """
        Alias for the adapter's read method

        :param args: any arguments for the adapter's read method
        :param kwargs: any arguments for the adapter's read method
        :return: whatever is returned by the adapter's read method
        """

        try:
            return self.adapter.read(*args, **kwargs)
        except AdapterError as adapter_error:
            if self.ignore_errors:
                logger.error(str(adapter_error))
            else:
                raise adapter_error

    def query(self, *args, **kwargs):
        """
        Alias for the adapter's query method, if it has one

        :param args: any arguments for the adapter's query method
        :param kwargs: any arguments for the adapter's query method
        :return: whatever is returned by the adapter's query method
        """

        try:
            return self.adapter.query(*args, **kwargs)
        except AdapterError as adapter_error:
            if self.ignore_errors:
                logger.error(str(adapter_error))
            else:
                raise adapter_error

    def set(self, knob: str, value):
        """
        Set the value of a knob on the instrument

        :param knob: (string) name of variable to be set
        :param value: (float/string) value of new variable setting
        :return: None
        """

        try:
            set_method = getattr(self, "set_" + knob.replace(" ", "_"))
        except AttributeError as err:
            raise AttributeError(f"{knob} cannot be set on {self.name}\n{err}")

        set_method(value)

    def get(self, knob: str):
        """
        Get the value of a knob on the instrument. If the instrument has a get
        method for the knob, a command will be sent to the instrument to
        retrieve the actual value of the knob. If it does not have a get method
        for the knob, the last known value, stored as an instance attribute,
        will be return (possibly being ``nan`` if no value has yet been set)

        """

        if hasattr(self, "get_" + knob.replace(" ", "_")):
            return getattr(self, "get_" + knob.replace(" ", "_"))()
        else:
            return getattr(self, knob.replace(" ", "_"))

    def measure(self, meter: str):
        """
        Measure the value of a variable associated with this instrument

        :param meter: (string) name of the variable to be measured
        :return: (float/string) measured value of the variable
        """

        try:
            measure_method = self.__getattribute__("measure_" + meter.replace(" ", "_"))
        except AttributeError:
            raise AttributeError(f"{meter} cannot be measured on {self.name}")

        measurement = measure_method()

        return measurement

    def connect(self):
        """
        (Re)Connect to the instrument. This is useful when communications are lost and
        a new connection is required.

        :return: None
        """

        try:
            if self.adapter.connected:
                self.adapter.disconnect()

            self.__init__(
                address=self.address,
                adapter=self.adapter if self.address is None else None,
                presets=self.presets,
                postsets=self.postsets,
                **self.kwargs,
            )
        except ConnectionError as connection_error:
            if self.ignore_errors:
                logger.error(str(connection_error))
            else:
                raise connection_error

    def disconnect(self):
        """
        Apply any postsets to the instrument and disconnect the adapter

        :return: None
        """

        if self.adapter.connected:
            for knob, value in self.postsets.items():
                try:
                    self.set(knob, value)
                except AdapterError as adapter_error:
                    if self.ignore_errors:
                        logger.error(str(adapter_error))
                    else:
                        raise adapter_error

            try:
                self.adapter.disconnect()
            except ConnectionError as connection_error:
                if self.ignore_errors:
                    logger.error(str(connection_error))
                else:
                    raise connection_error

        else:
            raise ConnectionError(f"adapter for {self.name} is not connected!")

    @setter
    def set_connected(self, state: Toggle):
        if state == ON:
            if self.adapter.connected:
                print(f"{self.name} is already connected")
            else:
                self.connect()
        elif state == OFF:
            if self.adapter.connected:
                self.disconnect()
            else:
                print(f"{self.name} is already disconnected")

    @getter
    def get_connected(self) -> Toggle:
        if self.adapter.connected:
            return ON
        else:
            return OFF



# Function/signal generators
from empyric.types import ON, OFF, Toggle, String, Float
from empyric.adapters import Socket
from empyric.instruments import Instrument, setter, getter


class SiglentSDG1000(Instrument):
    supported_adapters = (
        (Socket, {"write_termination": "\n", "read_termination": "\n"}),
    )

    knobs = (
        "channel 1 output",
        "channel 2 output",
        "channel 1 load",
        "channel 2 load",
        "channel 1 polarity",
        "channel 2 polarity",
        "channel 1 waveform",
        "channel 2 waveform",
        "channel 1 high level",
        "channel 2 high level",
        "channel 1 low level",
        "channel 2 low level",
        "channel 1 frequency",
        "channel 2 frequency",
        "channel 1 pulse width",
        "channel 2 pulse width",
        "channel 1 pulse rise",
        "channel 2 pulse rise",
        "channel 1 pulse fall",
        "channel 2 pulse fall",
        "channel 1 pulse delay",
        "channel 2 pulse delay",
        "channel 1 invert",
        "channel 2 invert",
        "channel 1 phase",
        "channel 2 phase",
        "equal phase",
        # BURST MODE PARAMETERS
        "channel 1 burst state",
        "channel 2 burst state",
        "channel 1 burst period",
        "channel 2 burst period",
        "channel 1 burst trigger source",
        "channel 2 burst trigger source",
        "channel 1 burst trigger delay",
        "channel 2 burst trigger delay",
        "channel 1 burst trigger edge",
        "channel 2 burst trigger edge",
        "channel 1 burst cycles",
        "channel 2 burst cycles",
        "channel 1 burst polarity",
        "channel 2 burst polarity",
    )

    wave_forms = ("SINE", "SQUARE", "RAMP", "PULSE", "NOISE", "ARB", "DC", "PRBS", "IQ")

    def _set_channel_n_output(self, n, output: Toggle, load: String, polarity: String):
        self.write(f"C{n}:OUTP {output},LOAD,{load},PLRT,{polarity}")

    def _get_channel_n_output(self, n) -> Toggle:
        response = self.query("C%d:OUTP?" % n).split(",")

        if response[0] == ("C%d:OUTP ON" % n):
            output = ON
        elif response[0] == ("C%d:OUTP OFF" % n):
            output = OFF
        else:
            raise ValueError(
                f"while trying to get output state of {self.name}, "
                f"got corrupted response {response}"
            )

        load = response[2]

        if response[4] == "NOR":
            polarity = "normal"
        elif response[4] == "INVT":
            polarity = "inverted"
        else:
            raise ValueError(
                f"while trying to get output state of {self.name}, "
                f"got corrupted response {response}"
            )

        return output, load, polarity

    def _set_channel_n_basic_waveform(self, n, **kwargs):
        waveform_dict = self._get_channel_n_waveform(n)

        for key in kwargs:
            if key.upper() not in waveform_dict:
                raise ValueError(
                    f"parameter {key} is not valid for waveform type "
                    f'{kwargs["WVTP"]}'
                )

        parameter_string = f"C{n}:BSWV " + ",".join(
            [f"{key.upper()},{value}" for key, value in kwargs.items()]
        )

        # Changing the waveform parameters can cause the relative phase
        # of the two channels to shift
        self.equal_phase = OFF

        self.write(parameter_string)

    def _set_channel_n_burst_waveform(self, n, **kwargs):
        waveform_dict = self._get_channel_n_waveform(n, mode="burst")

        if "wvtp" in kwargs:
            kwargs["CARR,WVTP"] = kwargs.pop("wvtp")

        for key in kwargs:
            if key.upper() not in waveform_dict:
                raise ValueError(f"parameter {key} is not valid for burst waveform")

        # For burst waveforms only, the wavetype parameter must be corrected
        parameter_string = f"C{n}:BTWV " + ",".join(
            [f"{key.upper()},{value}" for key, value in kwargs.items()]
        )

        # Changing the waveform parameters can cause the relative phase
        # of the two channels to shift
        self.equal_phase = OFF

        self.write(parameter_string)

    def _get_channel_n_waveform(self, n, mode="basic"):
        if mode.lower() == "burst":
            bt_response = self.query(f"C{n}:BTWV?")[3:]
            if "STATE,OFF" in bt_response:
                response = bt_response.split(f"BTWV ")[1].split(",")
            else:
                # Temporarily remove "carr" prefix to create waveform dictionary
                carrier_response = bt_response.split(f"BTWV ")[1].split("CARR,")[1]
                bt_response = bt_response.split(f"BTWV ")[1].split("CARR,")[0]
                response = (bt_response + carrier_response).split(",")
        else:
            bs_response = self.query(f"C{n}:BSWV?")
            response = bs_response.split(f"C{n}:BSWV ")[1].split(",")

        keys = response[::2]
        values = response[1::2]

        waveform_dict = {key: value for key, value in zip(keys, values)}

        # Adjust name of wavetype key for non-basic waveforms
        if mode.lower() != "basic" and waveform_dict["STATE"] != "OFF":
            waveform_dict["CARR,WVTP"] = waveform_dict.pop("WVTP")

        return waveform_dict

    def _set_channel_n_invert(self, n, state: Toggle):
        if state == ON:
            self.write("C%d:INVT ON" % n)
        elif state == OFF:
            self.write("C%d:INVT OFF" % n)

    def _get_channel_n_invert(self, n):
        response = self.query("C%d:INVT?" % n)

        if "ON" in response:
            return ON
        if "OFF" in response:
            return OFF

    # Output
    @setter
    def set_channel_1_output(self, output: Toggle):
        _, load, polarity = self._get_channel_n_output(1)

        self._set_channel_n_output(1, output, load, polarity)

    @getter
    def get_channel_1_output(self) -> Toggle:
        return self._get_channel_n_output(1)[0]

    @setter
    def set_channel_2_output(self, output: Toggle):
        _, load, polarity = self._get_channel_n_output(2)

        self._set_channel_n_output(2, output, load, polarity)

    @getter
    def get_channel_2_output(self) -> Toggle:
        return self._get_channel_n_output(2)[0]

    # Load
    @setter
    def set_channel_1_load(self, load: String):
        output, _, polarity = self._get_channel_n_output(1)

        self._set_channel_n_output(1, output, load, polarity)

    @getter
    def get_channel_1_load(self) -> String:
        return self._get_channel_n_output(1)[1]

    @setter
    def set_channel_2_load(self, load: String):
        output, _, polarity = self._get_channel_n_output(2)

        self._set_channel_n_output(2, output, load, polarity)

    @getter
    def get_channel_2_load(self) -> String:
        return self._get_channel_n_output(2)[1]

    # Polarity
    @setter
    def set_channel_1_polarity(self, polarity: String):
        output, load, _ = self._get_channel_n_output(1)

        self._set_channel_n_output(1, output, load, polarity)

    @getter
    def get_channel_1_polarity(self) -> String:
        return self._get_channel_n_output(1)[1]

    @setter
    def set_channel_2_polarity(self, polarity: String):
        output, load, _ = self._get_channel_n_output(2)

        self._set_channel_n_output(2, output, load, polarity)

    @getter
    def get_channel_2_polarity(self) -> String:
        return self._get_channel_n_output(2)[1]

    # Waveform type
    @setter
    def set_channel_1_waveform(self, waveform: String):
        if self.get_channel_1_burst_state() == ON:
            self._set_channel_n_burst_waveform(1, wvtp=waveform)
        else:
            self._set_channel_n_basic_waveform(1, wvtp=waveform)

    @getter
    def get_channel_1_waveform(self) -> String:
        if self.get_channel_1_burst_state() == ON:
            return self._get_channel_n_waveform(1, mode="burst")["CARR,WVTP"]
        else:
            return self._get_channel_n_waveform(1)["WVTP"]

    @setter
    def set_channel_2_waveform(self, waveform: String):
        if self.get_channel_2_burst_state() == ON:
            self._set_channel_n_burst_waveform(2, wvtp=waveform)
        else:
            self._set_channel_n_basic_waveform(2, wvtp=waveform)

    @getter
    def get_channel_2_waveform(self) -> String:
        if self.get_channel_2_burst_state() is ON:
            return self._get_channel_n_waveform(2, mode="burst")["CARR,WVTP"]
        else:
            return self._get_channel_n_waveform(2)["WVTP"]

    # Waveform high level
    @setter
    def set_channel_1_high_level(self, high_level: Float):
        self._set_channel_n_basic_waveform(1, hlev=f"{high_level}")

    @getter
    def get_channel_1_high_level(self) -> Float:
        high_level_str = self._get_channel_n_waveform(1).get("HLEV", "nan")

        return float(high_level_str.replace("V", ""))

    @setter
    def set_channel_2_high_level(self, high_level: Float):
        self._set_channel_n_basic_waveform(2, hlev=f"{high_level}")

    @getter
    def get_channel_2_high_level(self) -> Float:
        high_level_str = self._get_channel_n_waveform(2).get("HLEV", "nan")

        return float(high_level_str.replace("V", ""))

    # Waveform low level
    @setter
    def set_channel_1_low_level(self, low_level: Float):
        self._set_channel_n_basic_waveform(1, llev=f"{low_level}")

    @getter
    def get_channel_1_low_level(self) -> Float:
        low_level_str = self._get_channel_n_waveform(1).get("LLEV", "nan")

        return float(low_level_str.replace("V", ""))

    @setter
    def set_channel_2_low_level(self, low_level: Float):
        self._set_channel_n_basic_waveform(2, llev=f"{low_level}")

    @getter
    def get_channel_2_low_level(self) -> Float:
        low_level_str = self._get_channel_n_waveform(2).get("LLEV", "nan")

        return float(low_level_str.replace("V", ""))

    # Waveform frequency
    @setter
    def set_channel_1_frequency(self, frequency: Float):
        self._set_channel_n_basic_waveform(1, frq=f"{frequency}")

    @getter
    def get_channel_1_frequency(self) -> Float:
        freq_str = self._get_channel_n_waveform(1).get("FRQ", "nan")

        return float(freq_str.replace("HZ", ""))

    @setter
    def set_channel_2_frequency(self, frequency: Float):
        self._set_channel_n_basic_waveform(2, frq=f"{frequency}")

    @getter
    def get_channel_2_frequency(self) -> Float:
        freq_str = self._get_channel_n_waveform(2).get("FRQ", "nan")

        return float(freq_str.replace("HZ", ""))

    # Pulse width
    @setter
    def set_channel_1_pulse_width(self, width: Float):
        self._set_channel_n_basic_waveform(1, width=f"{width}")

    @getter
    def get_channel_1_pulse_width(self) -> Float:
        width_str = self._get_channel_n_waveform(1).get("WIDTH", "nan")

        return float(width_str.replace("S", ""))

    @setter
    def set_channel_2_pulse_width(self, width: Float):
        self._set_channel_n_basic_waveform(2, width=f"{width}")

    @getter
    def get_channel_2_pulse_width(self) -> Float:
        width_str = self._get_channel_n_waveform(2).get("WIDTH", "nan")

        return float(width_str.replace("S", ""))

    # Pulse rise
    @setter
    def set_channel_1_pulse_rise(self, rise: Float):
        self._set_channel_n_basic_waveform(1, rise=f"{rise}")

    @getter
    def get_channel_1_pulse_rise(self) -> Float:
        delay_str = self._get_channel_n_waveform(1).get("RISE", "nan")

        return float(delay_str.replace("S", ""))

    @setter
    def set_channel_2_pulse_rise(self, rise: Float):
        self._set_channel_n_basic_waveform(2, rise=f"{rise}")

    @getter
    def get_channel_2_pulse_rise(self) -> Float:
        delay_str = self._get_channel_n_waveform(2).get("RISE", "nan")

        return float(delay_str.replace("S", ""))

    # Pulse fall
    @setter
    def set_channel_1_pulse_fall(self, fall: Float):
        self._set_channel_n_basic_waveform(1, fall=f"{fall}")

    @getter
    def get_channel_1_pulse_fall(self) -> Float:
        delay_str = self._get_channel_n_waveform(1).get("FALL", "nan")

        return float(delay_str.replace("S", ""))

    @setter
    def set_channel_2_pulse_fall(self, fall: Float):
        self._set_channel_n_basic_waveform(2, fall=f"{fall}")

    @getter
    def get_channel_2_pulse_fall(self) -> Float:
        delay_str = self._get_channel_n_waveform(2).get("FALL", "nan")

        return float(delay_str.replace("S", ""))

    # Pulse delay
    @setter
    def set_channel_1_pulse_delay(self, delay: Float):
        self._set_channel_n_basic_waveform(1, dly=f"{delay}")

    @getter
    def get_channel_1_pulse_delay(self) -> Float:
        delay_str = self._get_channel_n_waveform(1).get("DLY", "nan")

        return float(delay_str.replace("S", ""))

    @setter
    def set_channel_2_pulse_delay(self, delay: Float):
        self._set_channel_n_basic_waveform(2, dly=f"{delay}")

    @getter
    def get_channel_2_pulse_delay(self) -> Float:
        delay_str = self._get_channel_n_waveform(2).get("DLY", "nan")

        return float(delay_str.replace("S", ""))

    # Output inversion
    @setter
    def set_channel_1_invert(self, state: Toggle):
        self._set_channel_n_invert(1, state)

    @getter
    def get_channel_1_invert(self) -> Toggle:
        return self._get_channel_n_invert(1)

    @setter
    def set_channel_2_invert(self, state: Toggle):
        self._set_channel_n_invert(2, state)

    @getter
    def get_channel_2_invert(self) -> Toggle:
        return self._get_channel_n_invert(2)
    
    # Phase (for SINE wave types)
    @setter
    def set_channel_1_phase(self, phase: Float):
        self._set_channel_n_basic_waveform(1, phse=f"{phase}")

    @getter
    def get_channel_1_phase(self) -> Float:
        delay_str = self._get_channel_n_waveform(1).get("PHSE", "nan")

        return float(delay_str)

    @setter
    def set_channel_2_phase(self, phase: Float):
        self._set_channel_n_basic_waveform(1, phse=f"{phase}")

    @getter
    def get_channel_2_phase(self) -> Float:
        delay_str = self._get_channel_n_waveform(2).get("PHSE", "nan")

        return float(delay_str)

    # Equalize phase of both channels
    @setter
    def set_equal_phase(self, state: Toggle):
        if state == ON:
            self.write("EQPHASE")

        # Equalizing the phase resets the delays, even if the display show otherwise.
        self.set_channel_1_pulse_delay(0.0)
        self.set_channel_2_pulse_delay(0.0)

        # One shot button
        return OFF

    # Burst activated
    @setter
    def set_channel_1_burst_state(self, state: Toggle):
        self._set_channel_n_burst_waveform(1, state=state)

    @getter
    def get_channel_1_burst_state(self) -> Toggle:
        return self._get_channel_n_waveform(1, mode="burst")["STATE"]

    @setter
    def set_channel_2_burst_state(self, state: Toggle):
        self._set_channel_n_burst_waveform(2, state=state)

    @getter
    def get_channel_2_burst_state(self) -> Toggle:
        return self._get_channel_n_waveform(2, mode="burst")["STATE"]

    # Burst period
    @setter
    def set_channel_1_burst_period(self, period: Float):
        self._set_channel_n_burst_waveform(1, prd=f"{period}")

    @getter
    def get_channel_1_burst_period(self) -> Float:
        prd_str = self._get_channel_n_waveform(1, mode="burst").get("PRD", "nan")

        return float(prd_str.replace("S", ""))

    @setter
    def set_channel_2_burst_period(self, period: Float):
        self._set_channel_n_burst_waveform(2, prd=f"{period}")

    @getter
    def get_channel_2_burst_period(self) -> Float:
        prd_str = self._get_channel_n_waveform(2, mode="burst").get("PRD", "nan")

        return float(prd_str.replace("S", ""))

    # Burst trigger source
    @setter
    def set_channel_1_burst_trigger_source(self, trsr: String):
        self._set_channel_n_burst_waveform(1, trsr=trsr)

    @getter
    def get_channel_1_burst_trigger_source(self) -> String:
        return self._get_channel_n_waveform(1, mode="burst").get("TRSR", None)

    @setter
    def set_channel_2_burst_trigger_source(self, trsr: String):
        self._set_channel_n_burst_waveform(2, trsr=trsr)

    @getter
    def get_channel_2_burst_trigger_source(self) -> String:
        return self._get_channel_n_waveform(2, mode="burst").get("TRSR", None)

    # Burst trigger delay
    @setter
    def set_channel_1_burst_trigger_delay(self, delay: Float):
        self._set_channel_n_burst_waveform(1, dlay=f"{delay}")

    @getter
    def get_channel_1_burst_trigger_delay(self) -> Float:
        dlay_str = self._get_channel_n_waveform(1, mode="burst").get("DLAY", "nan")

        return float(dlay_str.replace("S", ""))

    @setter
    def set_channel_2_burst_trigger_delay(self, delay: Float):
        self._set_channel_n_burst_waveform(2, dlay=f"{delay}")

    @getter
    def get_channel_2_burst_trigger_delay(self) -> Float:
        dlay_str = self._get_channel_n_waveform(2, mode="burst").get("DLAY", "nan")

        return float(dlay_str.replace("S", ""))

    # Burst trigger edge
    @setter
    def set_channel_1_burst_trigger_edge(self, edge: String):
        self._set_channel_n_burst_waveform(1, edge=edge)

    @getter
    def get_channel_1_burst_trigger_edge(self) -> String:
        return self._get_channel_n_waveform(1, mode="burst").get("EDGE", None)

    @setter
    def set_channel_2_burst_trigger_edge(self, edge: String):
        self._set_channel_n_burst_waveform(2, edge=edge)

    @getter
    def get_channel_2_burst_trigger_edge(self) -> String:
        return self._get_channel_n_waveform(2, mode="burst").get("EDGE", None)

    # Burst cycles
    @setter
    def set_channel_1_burst_cycles(self, cycles: Float):
        self._set_channel_n_burst_waveform(1, time=cycles)

    @getter
    def get_channel_1_burst_cycles(self) -> Float:
        return float(self._get_channel_n_waveform(1, mode="burst").get("TIME", "nan"))

    @setter
    def set_channel_2_burst_cycles(self, cycles: Float):
        self._set_channel_n_burst_waveform(2, time=cycles)

    @getter
    def get_channel_2_burst_cycles(self) -> Float:
        return float(self._get_channel_n_waveform(2, mode="burst").get("TIME", "nan"))

    # Burst polarity
    @setter
    def set_channel_1_burst_polarity(self, plrt: String):
        self._set_channel_n_burst_waveform(1, plrt=plrt)

    @getter
    def get_channel_1_burst_polarity(self) -> String:
        return self._get_channel_n_waveform(1, mode="burst").get("PLRT", None)

    @setter
    def set_channel_2_burst_polarity(self, plrt: String):
        self._set_channel_n_burst_waveform(2, plrt=plrt)

    @getter
    def get_channel_2_burst_polarity(self) -> String:
        return self._get_channel_n_waveform(2, mode="burst").get("PLRT", None)
